#!/usr/bin/python3
import numpy
import pathlib
import pandas
import xlsxwriter
import warnings

import logging
log = logging.getLogger(__name__)

from Bio import SeqIO

def _save_to_excel(data, output):
	#if overwrite:
	try:
		pathlib.Path(output).unlink(missing_ok = True) # python>=3.8
	except:
		if pathlib.Path(output).exists():
			pathlib.Path(output).unlink() # python==3.7

	with open(output, 'wb') as outfile:
		writer = pandas.ExcelWriter(outfile, engine = 'xlsxwriter')
		data.to_excel(writer, index = False, freeze_panes = (1, 8))
		(max_row, max_col) = data.shape

		# Get the xlsxwriter workbook and worksheet objects.
		workbook  = writer.book
		worksheet = writer.sheets['Sheet1']

		# Set the autofilter.
		worksheet.autofilter(0, 0, max_row, max_col - 1)

		# Make the columns wider for clarity.
		worksheet.set_column_pixels(0,  max_col - 1, 96)

		# Close the Pandas Excel writer and output the Excel file.
		writer.close()

def generate_organism_specific_matrix(genbank, locus_tag, model):
	contigs = []
	for contig in SeqIO.parse(genbank, 'genbank'):
		contigs.append(contig)

	# get all features
	lst = [ x for y in [ x.features for x in contigs ] for x in y ]
	lst = [ x for x in lst if x.type in [ 'CDS', 'rRNA', 'tRNA', 'ncRNA', 'tmRNA', 'misc_RNA' ] ]

	# create a pandas DataFrame with organism-specific information to be completed with the builder data
	df = pandas.DataFrame(columns = [
		'Gene Locus ID',
		'Definition',
		'Gene Names',
		'Old Locus Tag',
		'BioCyc',
		'Reference BBH',
		'Reference Definition',
		'Feature Type',
		'Complex Name',
		'Complex ID',
		'Cofactors in Modified Complex',
		'Generic Complex ID',
		'MetaComplex ID',
		'ME-model SubReaction',
		'M-model Reaction ID',
		'Reaction Name',
		'Reversibility',
		'GroEL_dependent_folding',
		'DnaK_dependent_folding',
		'N_terminal_methionine_cleavage',
		'RNA mods/enzyme',
		'tRNA-codon association',
		'RNA stability',
		'Complex Location',
		'Subunit Location',
		'Translocation Pathway',
		#'Translocation Multiplier'
	])

	def _get_reaction(x):
		if x is None:
			return []
		else:
			lst = []
			for gene in x.split(';'):
				if model.genes.has_id(gene):
					lst.append([ x.id for x in model.genes.get_by_id(gene).reactions ])

			# (An)A (old)locustag/gene name can be associated to many reactions
			lst = [ x for y in lst for x in y ]
			return lst

	def _get_reaction_name(x):
		if model.reactions.has_id(x):
			return model.reactions.get_by_id(x).name

	def _get_reversibility(x):
		if model.reactions.has_id(x):
			return 'True' if model.reactions.get_by_id(x).reversibility else 'False'

	def _get_spontaneous(x):
		if model.reactions.has_id(x):
			return 'True' if 'spontaneous' in model.reactions.get_by_id(x).name else 'False'

	#df['Gene Locus ID'] = [ x.qualifiers.get('locus_tag', [None])[0] for x in lst ]
	df['Gene Locus ID'] = [ x.qualifiers.get(locus_tag, [None])[0] for x in lst ]
	df['Definition'] = [ x.qualifiers.get('product', [None])[0] for x in lst ]
	df['Feature Type'] = [ x.type if x.qualifiers.get('pseudo') is None else 'pseudo' for x in lst ]

	tmp = [ x.qualifiers.get('gene', None) for x in lst ]
	df['Gene Names'] = [ ';'.join(x) if x is not None else None for x in tmp ]

	tmp = [ x.qualifiers['old_locus_tag'] if x.qualifiers.get('old_locus_tag', None) is not None else None for x in lst ]
	df['Old Locus Tag'] = [ ';'.join(x) if x is not None else None for x in tmp ]

	#df['M-model Reaction ID'] = df['Old Locus Tag'].apply(lambda x: _get_reaction(x))
	#df['M-model Reaction ID'] += df['Gene Names'].apply(lambda x: _get_reaction(x))
	#df['M-model Reaction ID'] += df['Gene Locus ID'].apply(lambda x: _get_reaction(x))
	#df = df.explode('M-model Reaction ID')

	#df['Reaction Name'] = df['M-model Reaction ID'].apply(lambda x: _get_reaction_name(x))
	#df['Reversibility'] = df['M-model Reaction ID'].apply(lambda x: _get_reversibility(x))

	# df.set_index(['Gene Locus ID', 'Definition', 'Feature type'], inplace = True)
	return df.sort_values(['M-model Reaction ID', 'Gene Locus ID'])

def complete_organism_specific_matrix(builder, data, model, output = False):
	if not hasattr(builder, 'org'):
		raise Exception('Please, run MEBuilder(*[configuration file]).generate_files() to generate the Organism-Specific Matrix.')

	# ME-model homology to reference
	def _get_homology(x, dct, keys):
		#tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'] ]
		tags = x[keys].to_list()
		tags = [ str(x).split(';') for x in tags ]

		lst = []
		for tag in [ x for y in tags for x in y ]:
			lst.append(dct.get(tag, None))
		lst = [ x for x in lst if x is not None ]

		if len(lst) != 0:
			return lst[0]

	def _get_definition(contigs, locus_tag):
		for contig in contigs:
			for feat in contig.features:
				if feat.type == 'CDS' and feat.qualifiers.get('locus_tag', [None])[0] == locus_tag:
					return ', '.join(feat.qualifiers.get('product', [None]))

	# Step 1a: Add reference data
	if hasattr(builder, 'homology'):
		dct = builder.homology.mutual_hits
		data['Reference BBH'] = data.apply(lambda x: _get_homology(x, dct, keys = ['Gene Locus ID', 'Old Locus Tag', 'BioCyc']), axis = 1)
		ref = builder.ref.contigs
		data['Reference Definition'] = data.apply(lambda x: _get_definition(ref, x['Reference BBH']), axis = 1)

	# Step 1b: Add BioCyc data if present
	if builder.configuration.get('biocyc.genes', False):
		# We reuse the _get_homology function, but changing the dictionary of relationships between IDs
		dct = { v['Accession-1']:k for k,v in builder.org.gene_dictionary.iterrows() }
		data['BioCyc'] = data.apply(lambda x: _get_homology(x, dct, keys = ['Gene Locus ID', 'Old Locus Tag', 'Gene Names']), axis = 1)
	else:
		data['BioCyc'] = None

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step1_get_homology'] = data.copy(deep = True)

	# Step 2: Add ME-model complexes
	# Restructure complexes_df to obtain the correct complex stoichiometry from the index
	df = builder.org.complexes_df.copy(deep = True)
	#df = df[~df.index.str.contains('MONOMER')]
	df = df[df['genes'].str.contains(r'\(')]
	df['genes'] = df['genes'].str.split(' AND ')
	df = df.explode('genes')
	df['stoich'] = df['genes'].apply(lambda x: '1' if x.split('(')[1][:-1] == '' else str(x.split('(')[1][:-1]))
	df['genes'] = df['genes'].apply(lambda x: x.split('(')[0].replace('-MONOMER', ''))
	df.index = df.index + ':' + df['stoich']

	def _get_complexes(x, df):
		tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'] ]
		tags = [ str(x).split(';') for x in tags ]

		cplx = []
		for tag in [ x for y in tags for x in y ]:
			res = df[df['genes'].str.fullmatch(tag)].index.to_list()
			if len(res) != 0:
				cplx.append(res)
		if len(cplx) != 0:
			return cplx[0]

	data['Complex ID'] = data.apply(lambda x: _get_complexes(x, df), axis = 1)
	data = data.explode('Complex ID')

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step2_get_complexes'] = data.copy(deep = True)

	# Step 3: Add ME-model cofactors
	def _get_cofactors(x, dct):
		tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], str(x['Complex ID']).split(':')[0] ]
		tags = [ str(x).split(';') for x in tags ]

		all_mods = []
		for tag in [ x for y in tags for x in y ]:
			all_mods.append([ x.split('_mod_')[1:] for x in dct.get(tag, [])]) # return a list of mods
		all_mods = [ x for y in all_mods for x in y ]
		# if multiple modifications: [['mn2(1)'], ['2fe2s(1)', 'mn2(1)'], ['4fe4s(1)', 'mn2(1)']]
		# into ['mn2(1)', '2fe2s(1) AND mn2(1), '4fe4s(1) AND mn2(1)]

		mod_strings = [None]
		if len(all_mods) != 0:
			for mods in all_mods:
				mods = [ x for x in mods if not x.startswith('SH') ] # Sulfur transfer in [enzyme]-S-sulfanylcysteine are enzymatic reactions
				logging.warning('The modification \'SH\' was removed. Add MetabolicReaction(s) to transfer sulfur from cysteine and to the many acceptors.')
				mods = [ x for x in mods if not x.startswith('Oxidized') ] # metabolic modification in ferredoxin and other proteins
				mods = [ x for x in mods if not x.startswith('palmitate') ] # metabolic modification from 2agpg160 in the lpp gene
				mods = [ x.replace('LI', 'li') for x in mods ]
				#mods = [ x.replace('lipo', 'lipoyl') for x in mods ]
				#logging.warning('The modification \'lipo\' was renamed to \'lipoyl\'.')
				#logging.warning('Add MetabolicReaction\'s to salvage or de novo synthesize lipoyl moieties. See https://www.genome.jp/pathway/map00785 for more information.')

				mods = [ x.replace('NiFeCoCN2', 'NiFe_cofactor') for x in mods ]
				logging.warning('The modification \'NiFeCoCN2\' was renamed to \'NiFe_cofactor\'.')
				logging.warning('Add MetabolicReactions to synthesize and transfer the NiFe_cofactor into the final acceptor enzyme. See https://biocyc.org/ECOLI/NEW-IMAGE?type=PATHWAY&object=PWY-8319 for more information.')

				if len(mods) != 0:
					if 'pan4p(1)' not in mods:
						mod_strings.append(' AND '.join(mods))
					else:
						# This remove all modification from holo-ACP protein. Modifications are metabolic.
						mod_strings.append('pan4p(1)')
						logging.warning('The modification \'pan4p\' was added. A MetabolicReaction will transfer pan4p from coenzyme A into Acyl-carrier protein(s).')

		if len(mod_strings) != 0:
			return mod_strings

	dct = {}
	for k,v in builder.org.protein_mod[['Core_enzyme']].to_dict()['Core_enzyme'].items():
		dct.setdefault(v, []).append(k)
	data['Cofactors in Modified Complex'] = data.apply(lambda x: _get_cofactors(x, dct), axis = 1)
	data = data.explode('Cofactors in Modified Complex')
	data = data.drop_duplicates()

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step3_get_cofactors'] = data.copy(deep = True)

	# Step 4: Add ME-model generics
	def _get_generics_from_gene(x, dct):
		generics = []
		tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Generic Complex ID'] ]
		tags = [ str(x).split(';') for x in tags ]
		for key, lst in dct.items():
			for tag in [ x for y in tags for x in y ]:
				if tag + '-MONOMER' == key[0]:
					generics.append(lst[0])
				if 'RNA_' + tag == key[0]:
					generics.append(lst[0])

		if len(generics) != 0:
			return [ x for y in generics for x in y ]

	def _get_generics_from_complex(x, dct):
		generics = []
		for key, lst in dct.items():
			if str(x['Complex ID']).split(':')[0] == key[0] and str(x['Cofactors in Modified Complex']).split(':')[0] == key[1]:
				generics.append(lst[0])
		if len(generics) != 0:
			return [ x for y in generics for x in y ]

	dct = pandas.DataFrame.from_dict({ k.replace('generic_', ''):v for k,v in builder.org.generic_dict.items() }).T
	dct = dct.explode('enzymes')
	dct['Complex'] = dct['enzymes'].apply(lambda gpr: [ x.split('_mod_')[0] for x in gpr.split(' OR ') ])
	dct['Cofactors'] = dct['enzymes'].apply(lambda gpr: [ ' AND '.join(x.split('_mod_')[1:]) for x in gpr.split(' OR ') ])
	dct['Generic'] = dct.index

	dct = dct.explode(['Complex', 'Cofactors']).replace('', 'None')
	dct = dct.groupby(['Complex', 'Cofactors']).agg({'Generic' : lambda x: set(x.tolist())})
	# final dictionary: (Complex, Cofactors) : Generic ID
	# we will repeat the Complex if it can form two or more generics
	#dct = { k:list(v['Generic'])[0] for k,v in dct.iterrows() }
	# final dictionary: (Complex, Cofactors) : [[Generic IDs]]
	tmp = {}
	for k,v in dct.iterrows():
		tmp.setdefault(k, []).append(list(v['Generic']))

	data['Generic Complex ID'] = data.apply(lambda x: _get_generics_from_gene(x, tmp), axis = 1)
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# data['Generic Complex ID'].update(data.apply(lambda x: _get_generics_from_complex(x, tmp), axis = 1))
	data.update({ 'Generic Complex ID' : data.apply(lambda x: _get_generics_from_complex(x, tmp), axis = 1) })
	data = data.explode('Generic Complex ID')

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step4_get_generics'] = data.copy(deep = True)

	# Step 5: Add M-model Reaction IDs, names, and reversibility from builder
	# It considers complex name and cofactors
	dct = builder.org.enz_rxn_assoc_df.copy(deep = True)
	dct['Complex'] = dct['Complexes'].apply(lambda gpr: [ x.split('_mod_')[0] for x in gpr.split(' OR ') ])
	dct['Cofactors'] = dct['Complexes'].apply(lambda gpr: [ ' AND '.join(x.split('_mod_')[1:]) for x in gpr.split(' OR ') ])
	dct['Reaction'] = dct.index

	dct = dct.explode(['Complex', 'Cofactors']).replace('', 'None')
	dct = dct.groupby(['Complex', 'Cofactors']).agg({'Reaction' : lambda x: x.tolist()})

	# intermediate dictionary: (Complex, Cofactors) : Reaction ID
	dct = { k:v['Reaction'] for k,v in dct.iterrows() }
	# correction based on generics -> NEW complex <-> list of reactions
	df = builder.org.complexes_df.copy(deep = True)
	df = { idx:[ x[8:-2] for x in row['genes'].split(' AND ')] for idx,row in df[df['genes'].str.contains('generic')].iterrows() }
	for cplx, generics in df.items():
		for generic in generics:
			dct.update({ (generic, 'None') : dct[(cplx, 'None')] })

	fn = lambda x: dct.get((str(x['Complex ID']).split(':')[0], str(x['Cofactors in Modified Complex'])), [None]) + dct.get((str(x['Generic Complex ID']).split(':')[0], str(x['Cofactors in Modified Complex'])), [None])

	data['M-model Reaction ID'] = data.apply(lambda x: fn(x), axis = 1)
	data = data.explode('M-model Reaction ID')
	data['Reaction Name'] = data['M-model Reaction ID'].apply(lambda x: model.reactions.get_by_id(x).name if model.reactions.has_id(x) else None)
	data['Reversibility'] = data['M-model Reaction ID'].apply(lambda x: str(model.reactions.get_by_id(x).reversibility) if model.reactions.has_id(x) else None)

	# correction based on generics -> NEW complex <-> list of reactions
	fn = lambda x: 'CPLX_{:s}-0:1({:s})'.format(x['M-model Reaction ID'], x['Generic Complex ID']) if x['M-model Reaction ID'] is not None and x['Generic Complex ID'] is not None else x['Complex ID']
	data['Complex ID'] = data.apply(lambda x: fn(x), axis = 1)
	fn = lambda x: None if x['M-model Reaction ID'] is not None else x['Generic Complex ID']
	data['Generic Complex ID'] = data.apply(lambda x: fn(x), axis = 1)
	data = data.drop_duplicates()

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step5_get_enzrxn_associations'] = data.copy(deep = True)

	# Step 6a: Add ME-model metacomplexes (e.g., ribosome)
	def _get_rnapol(x, lst):
		tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'] ]
		tags = [ str(x).split(';') for x in tags ]
		for tag in [ x for y in tags for x in y ]:
			if '{:s}-MONOMER'.format(tag) in lst or tag.split(':')[0] in lst:
				return 'RNAP'

	lst = builder.org.RNAP if isinstance(builder.org.RNAP, list) else [builder.org.RNAP]
	lst += list(builder.org.sigmas.index)
	data['MetaComplex ID'] = data.apply(lambda x: _get_rnapol(x, lst), axis = 1)

	def _get_ribosome(x, dct):
		mods = '' if x['Cofactors in Modified Complex'] is None else x['Cofactors in Modified Complex']
		#tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'], x['Generic Complex ID'] ]
		tags = [ x['Complex ID'], x['Generic Complex ID'] ]
		tags = [ str(x).split(';') for x in tags ]
		for tag in [ x for y in tags for x in y ]:
			#if '{:s}-MONOMER'.format(tag) in lst or tag.split(':')[0] in lst or 'generic_{:s}'.format(tag) in lst:
			filter1 = '{:s}-MONOMER'.format(tag) in dct and mods in dct.get('{:s}-MONOMER'.format(tag), [None])
			filter2 = tag.split(':')[0] in dct and mods in dct.get(tag.split(':')[0], [None])
			filter3 = 'generic_{:s}'.format(tag) in dct #and mods in dct.get('generic_{:s}'.format(tag), [None])

			if (filter1 or filter2 or filter3) and mods == '':
				return 'ribosome:1'
			elif (filter1 or filter2 or filter3) and mods == 'acetyl(1)':
				return 'ribosome:2'

	dct = {}
	for x in [ x for y in [builder.org.ribosome_stoich[x]['stoich'].keys() for x in ['30_S_assembly', '50_S_assembly']] for x in y ]:
		dct.setdefault(x.split('_mod_')[0], []).append(' AND '.join(x.split('_mod_')[1:]))
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# data['MetaComplex ID'].update(data.apply(lambda x: _get_ribosome(x, dct), axis = 1))
	data.update({ 'MetaComplex ID' : data.apply(lambda x: _get_ribosome(x, dct), axis = 1) })

	def _get_degradosome_and_excision(x, dct):
		subrxns = []
		for key, subdct in dct.items():
			mods = '' if x['Cofactors in Modified Complex'] is None else x['Cofactors in Modified Complex']
			#tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'], x['Generic Complex ID'] ]
			tags = [ x['Complex ID'], x['Generic Complex ID'] ]
			tags = [ str(x).split(';') for x in tags ]
			for tag in [ x for y in tags for x in y ]:
				#if '{:s}-MONOMER'.format(tag) in lst or tag.split(':')[0] in lst or 'generic_{:s}'.format(tag) in lst:
				filter1 = '{:s}-MONOMER'.format(tag) in subdct and mods in subdct.get('{:s}-MONOMER'.format(tag), [None])
				filter2 = tag.split(':')[0] in subdct and mods in subdct.get(tag.split(':')[0], [None])
				filter3 = 'generic_{:s}'.format(tag) in subdct #and mods in subdct.get('generic_{:s}'.format(tag), [None])

				if filter1 or filter2 or filter3:
					subrxns.append(key + ':1')

		if len(subrxns) != 0:
			return subrxns

	dct = builder.org.excision_machinery
	dct['RNA_degradosome'] = builder.org.rna_degradosome['rna_degradosome']

	dct = { k:[ x for x in v['enzymes'] ] for k,v in dct.items() }
	for key in dct:
		tmp = {}
		for x in dct[key]:
			tmp.setdefault(x.split('_mod_')[0], []).append(' AND '.join(x.split('_mod_')[1:]))
		dct[key] = tmp

	# Copy-on-Write pandas 3.0 behavior is default to be True
	# data['MetaComplex ID'].update(data.apply(lambda x: _get_degradosome_and_excision(x, dct), axis = 1))
	data.update({ 'MetaComplex ID' : data.apply(lambda x: _get_degradosome_and_excision(x, dct), axis = 1) })

	# Step 6b: set RNA modifiers (CDS<->mod_at_position). Part1: Add 'RNA_modifier_enzyme' in 'MetaComplex ID' column
	def _get_rna_modifiers(x, dct):
		mods = '' if x['Cofactors in Modified Complex'] is None else x['Cofactors in Modified Complex']
		tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'], x['Generic Complex ID'] ]
		tags = [ str(x).split(';') for x in tags ]
		for tag in [ x for y in tags for x in y ]:
			filter1 = '{:s}-MONOMER'.format(tag) in dct and mods in dct.get('{:s}-MONOMER'.format(tag), [None])
			filter2 = tag.split(':')[0] in dct and mods in dct.get(tag.split(':')[0], [None])
			filter3 = 'generic_{:s}'.format(tag) in dct #and mods in dct.get('generic_{:s}'.format(tag), [None])

			if filter1 or filter2 or filter3:
				return 'RNA_modifier_enzyme'

	dct = { k.split('_mod_')[0]:' AND '.join(k.split('_mod_')[1:]) for k,v in builder.org.rna_modification.items() }
	dct = { k:v for k,v in dct.items() } # if v != '' } # just in case of empty values
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# data['MetaComplex ID'].update(data.apply(lambda x: _get_rna_modifiers(x, dct), axis = 1))
	data.update({ 'MetaComplex ID' : data.apply(lambda x: _get_rna_modifiers(x, dct), axis = 1) })

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step6_get_metacomplexes'] = data.copy(deep = True)

	# Step 7a: set RNA targets (tRNA<->mod_at_position)
	dct = builder.org.rna_modification_targets.copy(deep = True)
	dct['at'] = dct['modification'] + '_at_' + dct['position'].astype(str)
	dct = dct.groupby('bnum').agg({'at': lambda x: ','.join(x.tolist())}).to_dict()['at']
	data['RNA mods/enzyme'] = data['Gene Locus ID'].apply(lambda x: dct.get(str(x), None))

	# Step 7b: set RNA modifiers (CDS<->mod_at_position). Part2: Add 'modification' list in 'RNA mods/enzyme' column
	def _get_rna_modifications(x, dct):
		mod_at_pos = set()
		mods = '' if x['Cofactors in Modified Complex'] is None else x['Cofactors in Modified Complex']
		tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'], x['Generic Complex ID'] ]
		tags = [ str(x).split(';') for x in tags ]
		for tag in [ x for y in tags for x in y ]:
			filter1 = ('{:s}-MONOMER'.format(tag), mods) in dct
			filter2 = (tag.split(':')[0], mods) in dct
			filter3 = ('generic_{:s}'.format(tag), '') in dct

			if filter1 or filter2 or filter3:
				mod_at_pos.add(dct.get(('{:s}-MONOMER'.format(tag), mods), None))
				mod_at_pos.add(dct.get((tag.split(':')[0], mods), None))
				mod_at_pos.add(dct.get(('generic_{:s}'.format(tag), ''), None))

		lst = list(set([ x for x in mod_at_pos if x is not None ]))
		if len(lst) != 0:
			return ','.join(lst)

	dct = { (k.split('_mod_')[0], ' AND '.join(k.split('_mod_')[1:])):','.join(v) for k,v in builder.org.rna_modification.items() }
	dct = { k:v for k,v in dct.items() } # if v != '' } # just in case of empty values
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# data['RNA mods/enzyme'].update(data.apply(lambda x: _get_rna_modifications(x, dct), axis = 1))
	data.update({ 'RNA mods/enzyme' : data.apply(lambda x: _get_rna_modifications(x, dct), axis = 1) })

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step7_get_trna_mods_and_modifiers'] = data.copy(deep = True)

	# Step 8: Add translocation pathways
	def _get_transpaths(x, dct):
		pathways = []
		for key, lst in dct.items():
			tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'], x['Generic Complex ID'] ]
			tags = [ str(x).split(';') for x in tags ]
			for tag in [ x for y in tags for x in y ]:
				if '{:s}-MONOMER'.format(tag) in lst or tag.split(':')[0] in lst or 'generic_{:s}'.format(tag) in lst:
					pathways.append('translocation_pathway_' + key)
		if len(pathways) != 0:
			return pathways

	dct = { k:v['enzymes'] for k,v in builder.org.translocation_pathways.items() if len(v['enzymes']) != 0 }
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# data['MetaComplex ID'].update(data.apply(lambda x: _get_transpaths(x, dct), axis = 1))
	data.update({ 'MetaComplex ID' : data.apply(lambda x: _get_transpaths(x, dct), axis = 1) })
	data = data.explode('MetaComplex ID')

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step8_get_translocation_pathways'] = data.copy(deep = True)

	# Step 9: Add ME-model subreactions
	# TODO: generalize to complexes with cofactors
	def _get_ribosome_subrxns(x, dct):
		for key, lst in dct.items():
			mods = '' if x['Cofactors in Modified Complex'] is None else x['Cofactors in Modified Complex']
			tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'], x['Generic Complex ID'] ]
			tags = [ str(x).split(';') for x in tags ]
			for tag in [ x for y in tags for x in y ]:
				if '{:s}-MONOMER'.format(tag) in lst or tag.split(':')[0] in lst or 'generic_{:s}'.format(tag) in lst:
					return 'Ribosome_' + key if 'Ribosome_' not in key else key

	dct = { k:[ x.split('_mod_')[0] for x in [v['enzyme']] ] for k,v in builder.org.ribosome_subreactions.items() }
	data['ME-model SubReaction'] = data.apply(lambda x: _get_ribosome_subrxns(x, dct), axis = 1)

	def _get_translation_subrxns(x, dct):
		subprocess = []
		for key, subdct in dct.items():
			mods = '' if x['Cofactors in Modified Complex'] is None else x['Cofactors in Modified Complex']
			tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'], x['Generic Complex ID'] ]
			tags = [ str(x).split(';') for x in tags ]
			for tag in [ x for y in tags for x in y ]:
				filter1 = '{:s}-MONOMER'.format(tag) in subdct and mods in subdct.get('{:s}-MONOMER'.format(tag), [None])
				filter2 = tag.split(':')[0] in subdct and mods in subdct.get(tag.split(':')[0], [None])
				filter3 = 'generic_{:s}'.format(tag) in subdct #and mods in subdct.get('generic_{:s}'.format(tag), [None])

				if filter1 or filter2 or filter3:
					if key.endswith('InfA') or key.endswith('InfC'):
						subprocess.append(key)
					elif key.endswith('InfB'):
						subprocess.append('Translation_initiation_gtp_factor_InfB')
					elif key == 'fmet_addition_at_START':
						subprocess.append('Translation_initiation_' + key)
					elif key == 'FusA_mono_elongation':
						subprocess.append('Translation_elongation_FusA_mono')
					elif key == 'Tuf_gtp_regeneration':
						subprocess.append('Translation_elongation_' + key)
					elif key in ['N_terminal_methionine_cleavage', 'DnaK_dependent_folding', 'GroEL_dependent_folding']:
						subprocess.append('Protein_processing_' + key)
					elif key == 'PrfA_mono_mediated_termination':
						subprocess.append('Translation_termination_PrfA_mono_mediated')
					elif key == 'PrfB_mono_mediated_termination':
						subprocess.append('Translation_termination_PrfB_mono_mediated')
					elif key == 'generic_RF_mediated_termination':
						subprocess.append('Translation_termination_generic_RF_mediated')
					else:
						subprocess.append('Translation_termination_' + key)

		if len(subprocess) != 0:
			return subprocess

	dct = { k:[ x for x in v['enzymes'] ] for k,v in builder.org.initiation_subreactions.items() }
	dct.update({ k:[ x for x in v['enzymes'] ] for k,v in builder.org.elongation_subreactions.items() })
	dct.update({ k:[ x for x in v['enzymes'] ] for k,v in builder.org.termination_subreactions.items() })

	for key in dct:
		tmp = {}
		for x in dct[key]:
			tmp.setdefault(x.split('_mod_')[0], []).append(' AND '.join(x.split('_mod_')[1:]))
		dct[key] = tmp

	# Copy-on-Write pandas 3.0 behavior is default to be True
	# data['ME-model SubReaction'].update(data.apply(lambda x: _get_translation_subrxns(x, dct), axis = 1))
	data.update({ 'ME-model SubReaction' : data.apply(lambda x: _get_translation_subrxns(x, dct), axis = 1) })
	data = data.explode('ME-model SubReaction')

	def _get_transcription_subrxns(x, dct):
		subrxns = []
		for key, subdct in dct.items():
			mods = '' if x['Cofactors in Modified Complex'] is None else x['Cofactors in Modified Complex']
			tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Complex ID'], x['Generic Complex ID'] ]
			tags = [ str(x).split(';') for x in tags ]
			for tag in [ x for y in tags for x in y ]:
				filter1 = '{:s}-MONOMER'.format(tag) in subdct and mods in subdct.get('{:s}-MONOMER'.format(tag), [None])
				filter2 = tag.split(':')[0] in subdct and mods in subdct.get(tag.split(':')[0], [None])
				filter3 = 'generic_{:s}'.format(tag) in subdct #and mods in subdct.get('generic_{:s}'.format(tag), [None])

				if filter1 or filter2 or filter3:
					subrxns.append(key)
		if len(subrxns) != 0:
			return subrxns

	dct = { k:[ x for x in v['enzymes'] ] for k,v in builder.org.transcription_subreactions.items() }
	for key in dct:
		tmp = {}
		for x in dct[key]:
			tmp.setdefault(x.split('_mod_')[0], []).append(' AND '.join(x.split('_mod_')[1:]))
		dct[key] = tmp

	# Copy-on-Write pandas 3.0 behavior is default to be True
	# data['ME-model SubReaction'].update(data.apply(lambda x: _get_transcription_subrxns(x, dct), axis = 1))
	data.update({ 'ME-model SubReaction' : data.apply(lambda x: _get_transcription_subrxns(x, dct), axis = 1) })
	data = data.explode('ME-model SubReaction')

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step9_get_subreactions'] = data.copy(deep = True)

	# Step 10: Add post-translation processing
	def _get_processing_targets(x, lst):
		tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'] ]
		tags = [ str(x).split(';') for x in tags ]
		for tag in [ x for y in tags for x in y ]:
			if tag in lst:
				return 'True'

	lst = builder.org.folding_dict['GroEL_dependent_folding']['enzymes']
	data['GroEL_dependent_folding'] = data.apply(lambda x: _get_processing_targets(x, lst), axis = 1)
	lst = builder.org.folding_dict['DnaK_dependent_folding']['enzymes']
	data['DnaK_dependent_folding'] = data.apply(lambda x: _get_processing_targets(x, lst), axis = 1)
	lst = builder.org.cleaved_methionine
	data['N_terminal_methionine_cleavage'] = data.apply(lambda x: _get_processing_targets(x, lst), axis = 1)
# 	lst = builder.org.stable_RNAs
# 	data['RNA stability'] = data.apply(lambda x: _get_processing_targets(x, lst), axis = 1)

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step10_get_posttranslation_processing'] = data.copy(deep = True)

	# Step 11: Add tRNA to codon association from GenBank data
	dct = { k:','.join(v) for x in [ v for k,v in builder.me_model.global_info['trna_to_codon'].items() ] for k,v in x.items() }
	data['tRNA-codon association'] = data.apply(lambda x: dct.get(x['Gene Locus ID'], None), axis = 1)

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step11_get_trna2codon_specificity'] = data.copy(deep = True)

	# Step 12: protein location
	def _get_protein_location(x, df):
		tags = [ x['Gene Locus ID'], x['Old Locus Tag'], x['BioCyc'], x['Generic Complex ID'] ]
		tags = [ str(x).split(';') for x in tags ] # we convert here None to 'None'
		tags = [ x for y in tags for x in y ]
		# extra modification if compared this function to the others above
		tags = [ r'{:s}(\(\d*\)|\(\d*\:\d*\))'.format(x) for x in tags if x != 'None' ]

		#res = df[df['Protein'].str.match(x)][['Complex_compartment', 'Protein_compartment', 'translocase_pathway']].values
		#query = [ '{:s}()'.format(y) for y in tags ] # sp old_locus_tag's, the query can be a string of tags separated by ';'
		res = df[df['Protein'].str.fullmatch('|'.join(tags))][['Complex_compartment', 'Protein_compartment', 'translocase_pathway']].values

		if len(res) != 0:
			return res[0]
		else:
			return [None, None, None]

	df = builder.org.protein_location.dropna(how = 'all', subset = ['Complex_compartment', 'Protein', 'Protein_compartment'])
	data['Complex Location'], data['Subunit Location'], data['Translocation Pathway'] = zip(*data.apply(lambda x: _get_protein_location(x, df), axis = 1))
	data = data.explode('Complex Location')

	# debug purposes
	if builder.configuration.get('collect_debug_data', False):
		builder.debug_data['complete_organism_specific_matrix_step12_get_protein_location'] = data.copy(deep = True)

	# final sorting
	data = data.sort_values(['M-model Reaction ID', 'Gene Locus ID'])
	data = data.drop_duplicates()

	if output:
		try:
			# Save file as excel or tsv depending on the size
			if not str(output).endswith('.xlsx'):
				output = '.'.join(output.split('.')[:-1]) + '.xlsx'

			# GPRs can expand the model specifications beyond the max size of an excel file (1048576 rows)
			if data.shape[0] > 1048575: # one less to accommodate the header
				# Divide the DataFrame into pieces and save them
				rxn_ids = data.groupby(['M-model Reaction ID'], dropna = True).size() # Anything with 1 or more occurrences

				gene_ids = data[data['M-model Reaction ID'].isin(rxn_ids[rxn_ids < 1000].index)]['Gene Locus ID']
				tmp = pandas.concat([ data[data['Gene Locus ID'].isin(gene_ids)], data[data['M-model Reaction ID'].isna()]], axis = 1)
				_save_to_excel(tmp, '.'.join(output.split('.')[:-1]) + '_{:02d}.xlsx'.format(0))

				if output.endswith('.xlsx'):
					for idx, gene_id in enumerate(rxn_ids[rxn_ids >= 1000].index):
						tmp = data[data['Gene Locus ID'].isin([gene_id])]
						_save_to_excel(tmp, '.'.join(output.split('.')[:-1]) + '_{:03d}.xlsx'.format(idx+1))

				# save the output as a tsv file
				if not output.endswith('.txt'):
					output = '.'.join(output.split('.')[:-2]) + '.txt'

				with open(output, 'w') as outfile:
					data.to_csv(outfile, index = False, sep = '\t')
			else:
				_save_to_excel(data, output)
		except:
			logging.warning('The builder.df_data was not saved to the \'{:s}\' file.'.format(output))

	return data

def _correct_input(df):
	# correct Gene Locus ID to reflect if they are proteins or RNAs
	fn = lambda x: \
		'protein_{:s}'.format(x['Gene Locus ID']) if (x['Feature Type'] == 'CDS' and not x['Gene Locus ID'].startswith('protein_')) \
		else 'RNA_{:s}'.format(x['Gene Locus ID']) if (x['Feature Type'] != 'pseudo' and not x['Gene Locus ID'].startswith('RNA_')) \
		else 'pseudo_{:s}'.format(x['Gene Locus ID'])
	df.loc[:, 'Gene Locus ID'] = df[['Gene Locus ID', 'Feature Type']].apply(fn, axis = 1)

	# correct generics to contain the `generic_` prefix
	fn = lambda x: 'generic_{:s}'.format(x) if isinstance(x, str) and not x.startswith('generic_') else x
	df.loc[:, 'Generic Complex ID'] = df['Generic Complex ID'].apply(fn)
	return df

# 2nd step
def get_generics(df):
	#tmp = df[df['Generic Complex ID'].notna() & ~df['Feature Type'].isin(['pseudo'])]
	tmp = df[df['Generic Complex ID'].notna()].copy() # this was a view
	tmp = _correct_input(tmp)

	fn = lambda x: x['Complex ID'].split(':')[0] + ''.join([ '_mod_{:s}'.format(x) for x in x['Cofactors in Modified Complex'].split(' AND ')]) \
		if isinstance(x['Cofactors in Modified Complex'], str) else numpy.nan
	tmp.loc[:, 'Modified Complex'] = tmp.apply(fn, axis = 1).values

	# collapse
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# tmp['Complex ID'].update(tmp['Modified Complex']) # inplace
	tmp.update({ 'Complex ID' : tmp['Modified Complex'] })
	# tmp['Gene Locus ID'].update(tmp['Complex ID']) # inplace
	tmp.update({ 'Gene Locus ID' : tmp['Complex ID'] })

	tmp.loc[:, 'Gene Locus ID'] = tmp['Gene Locus ID'].apply(lambda x: x.split(':')[0] if isinstance(x, str) else x)
	tmp = tmp.groupby(['Generic Complex ID']).agg({'Gene Locus ID': lambda x: sorted(set(x.tolist()))}).to_dict()['Gene Locus ID'].items()
	return tmp

def _get_metacomplex_stoichiometry(df, key):
	#tmp = df[df['MetaComplex ID'].notna() & df['MetaComplex ID'].str.startswith(key) & ~df['Feature Type'].isin(['pseudo'])]
	tmp = df[df['MetaComplex ID'].notna() & df['MetaComplex ID'].str.startswith(key)].copy() # this was a view
	if not tmp.empty:
		tmp = _correct_input(tmp)

		tmp.loc[:, 'Complex ID'] = tmp['Complex ID'].apply(lambda x: x.split(':')[0] if isinstance(x, str) else x)

		fn = lambda x: x['Complex ID'].split(':')[0] + ''.join([ '_mod_{:s}'.format(x) for x in x['Cofactors in Modified Complex'].split(' AND ')]) \
			if isinstance(x['Cofactors in Modified Complex'], str) else None #numpy.nan
		tmp.loc[:, 'Modified Complex'] = tmp[['Complex ID', 'Cofactors in Modified Complex']].apply(fn, axis = 1)

		# collapse
		# Copy-on-Write pandas 3.0 behavior is default to be True
		# tmp['Modified Complex'].update(tmp['Generic Complex ID']) # inplace
		tmp.update({ 'Modified Complex' : tmp['Generic Complex ID'] })
		# tmp['Complex ID'].update(tmp['Modified Complex']) # inplace
		tmp.update({ 'Complex ID' : tmp['Modified Complex'] })
		# tmp['Gene Locus ID'].update(tmp['Complex ID']) # inplace
		tmp.update({ 'Gene Locus ID' : tmp['Complex ID'] })

		tmp.loc[:, 'stoich'] = tmp['MetaComplex ID'].str.split(':', expand = True).iloc[:, 1]
		return tmp

def ribosome_stoichiometry(df):
	tmp = _get_metacomplex_stoichiometry(df, 'ribosome')
	ribosome_stoich = { 'ribosome' : { 'stoich' : { k.split(':')[0]:int(v) for k,v in zip(tmp['Gene Locus ID'], tmp['stoich'])}}}
	ribosome_stoich['ribosome']['stoich']['gtp_c'] = 1
	return ribosome_stoich

def degradosome_stoichiometry(df):
	tmp = _get_metacomplex_stoichiometry(df, 'RNA_degradosome')
	return { k.split(':')[0]:int(v) for k,v in zip(tmp['Gene Locus ID'], tmp['stoich']) } if tmp is not None else None

def dnapolymerase_stoichiometry(df):
	tmp = _get_metacomplex_stoichiometry(df, 'DNAP')
	return { k.split(':')[0]:int(v) for k,v in zip(tmp['Gene Locus ID'], tmp['stoich']) }

def excision_machinery_stoichiometry(df, keys):
	#tmp = df[df['MetaComplex ID'].notna() & df['MetaComplex ID'].str.match(keys) & ~df['Feature Type'].isin(['pseudo'])]
	tmp = df[df['MetaComplex ID'].notna() & df['MetaComplex ID'].str.match(keys)].copy() # this was a view
	if not tmp.empty:
		tmp = _correct_input(tmp)

		fn = lambda x: x['Complex ID'].split(':')[0] + ''.join([ '_mod_{:s}'.format(x) for x in x['Cofactors in Modified Complex'].split(' AND ')]) \
			if isinstance(x['Cofactors in Modified Complex'], str) else numpy.nan
		tmp.loc[:, 'Modified Complex'] = tmp.apply(fn, axis = 1)

		# collapse
		# Copy-on-Write pandas 3.0 behavior is default to be True
		# tmp['Modified Complex'].update(tmp['Generic Complex ID']) # inplace
		tmp.update({ 'Modified Complex' : tmp['Generic Complex ID'] })
		# tmp['Complex ID'].update(tmp['Modified Complex']) # inplace
		tmp.update({ 'Complex ID' : tmp['Modified Complex'] })
		# tmp['Gene Locus ID'].update(tmp['Complex ID']) # inplace
		tmp.update({ 'Gene Locus ID' : tmp['Complex ID'] })

		tmp.loc[:, 'stoich'] = tmp['MetaComplex ID'].str.split(':', expand = True).iloc[:, 1]
		return { k.split(':')[0]:int(v) for k,v in zip(tmp['Gene Locus ID'], tmp['stoich']) }

def aa_synthetase_dict(df):
	#tmp = df[df['Definition'].str.contains('--tRNA ligase|-tRNA synthetase') & ~df['Feature Type'].isin(['pseudo'])]
	tmp = df[df['Definition'].str.contains('--tRNA ligase|-tRNA synthetase') & df['Definition'].notna()]
	tmp = _correct_input(tmp)

	fn = lambda x: x['Complex ID'].split(':')[0] + ''.join([ '_mod_{:s}'.format(x) for x in x['Cofactors in Modified Complex'].split(' AND ')]) \
		if isinstance(x['Cofactors in Modified Complex'], str) else numpy.nan
	tmp['Modified Complex'] = tmp.apply(fn, axis = 1)

	# collapse
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# tmp['Modified Complex'].update(tmp['Generic Complex ID']) # inplace
	tmp.update({ 'Modified Complex' : tmp['Generic Complex ID'] })
	# tmp['Complex ID'].update(tmp['Modified Complex']) # inplace
	tmp.update({ 'Complex ID' : tmp['Modified Complex'] })
	# tmp['Gene Locus ID'].update(tmp['Complex ID']) # inplace
	tmp.update({ 'Gene Locus ID' : tmp['Complex ID'] })

	tmp = tmp.drop_duplicates(['Gene Locus ID'])

	tmp = { k.split('--tRNA')[0].lower():v.split(':')[0] for k,v in zip(tmp['Definition'], tmp['Gene Locus ID']) }
	tmp = { k.replace('asparagine', 'asn'):v for k,v in tmp.items() }
	tmp = { k.replace('isoleucine', 'ile'):v for k,v in tmp.items() }
	tmp = { k.replace('tryptophan', 'trp'):v for k,v in tmp.items() }
	return { (k[:3]+'__L_c').replace('gly__L_c', 'gly_c'):v for k,v in tmp.items() }

def get_subreactions(df, key: str):
	tmp = df.copy(deep = True)
	#tmp = tmp[~tmp['Feature Type'].isin(['pseudo'])]
	tmp = _correct_input(tmp)
	tmp = tmp[tmp['ME-model SubReaction'].notna()]

	tmp['Gene Locus ID'] = tmp['Gene Locus ID'].apply(lambda x: '{:s}_cplx'.format(x))
	tmp['Complex ID'] = tmp['Complex ID'].apply(lambda x: x.split(':')[0] if isinstance(x, str) else x)

	fn = lambda x: x['Complex ID'].split(':')[0] + ''.join([ '_mod_{:s}'.format(x) for x in x['Cofactors in Modified Complex'].split(' AND ')]) \
		if isinstance(x['Cofactors in Modified Complex'], str) else numpy.nan
	tmp['Modified Complex'] = tmp[['Complex ID', 'Cofactors in Modified Complex']].apply(fn, axis = 1)

	# collapse
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# tmp['Modified Complex'].update(tmp['Generic Complex ID']) # inplace
	tmp.update({ 'Modified Complex' : tmp['Generic Complex ID'] })
	# tmp['Complex ID'].update(tmp['Modified Complex']) # inplace
	tmp.update({ 'Complex ID' : tmp['Modified Complex'] })
	# tmp['Gene Locus ID'].update(tmp['Complex ID']) # inplace
	tmp.update({ 'Gene Locus ID' : tmp['Complex ID'] })

	# Correct Translation_termination_generic_RF_mediated
	if key.startswith('Translation_termination_generic_RF_mediated'):
		cplxs = tmp[tmp['ME-model SubReaction'].str.match('Translation_termination_generic_RF_mediated')]['Complex ID'].tolist()
		cplxs = set([ x.split(':')[0] for x in cplxs ])

	tmp = tmp.groupby(['ME-model SubReaction']).agg({'Gene Locus ID': lambda x: x.tolist()})
	if key.startswith('Translation_termination_generic_RF_mediated'):
		tmp = tmp.apply(lambda x: [['generic_RF']] if set(x['Gene Locus ID']) == cplxs else x, axis = 1)

	return { k : {'enzymes' : list(set(v)), 'stoich': {}, 'element_contribution' : {}, 'keff' : []}
		for k,v in zip(tmp.index, tmp['Gene Locus ID']) if k.startswith(key) }

def _get_df_rxns(df):
	tmp = df.copy(deep = True)
	#tmp = tmp[~tmp['Feature Type'].isin(['pseudo'])]
	tmp = tmp[tmp['M-model Reaction ID'].notna()]
	tmp = tmp[['M-model Reaction ID', 'Reaction Name', 'Reversibility']]
	tmp.columns = ['name', 'description', 'is_reversible']
	tmp = tmp.dropna().drop_duplicates('name', keep = 'first').set_index('name')

	return tmp

def _get_df_cplxs(df, generics = False):
	tmp = df[df['Feature Type'].isin(['CDS', 'pseudo', 'ncRNA'])].fillna('')

	# get enzymatic complexes with generics subunits
	df_generics = df[df['Feature Type'].isin(['CDS'])].fillna('')
	df_generics = df_generics[df_generics['Complex ID'].str.contains(r'\(')]
	df_generics['Gene Locus ID'] = 'generic_' + df_generics['Complex ID'].str.extract(r'[A-Za-z0-9\_-]+:\d+\(([A-Za-z0-9\_-]+)\)', expand = True)
	df_generics['Complex ID'] = df_generics['Complex ID'].str.extract(r'([A-Za-z0-9\_-]+:\d+)\([A-Za-z0-9\_-]+\)', expand = True)

	tmp = pandas.concat([tmp, df_generics])
	tmp = tmp[~tmp['Complex ID'].str.contains(r'\(')]

	fn = lambda x: '{:s} complex'.format(x['Definition']) if x['Complex Name'] == '' else x['Complex Name']
	tmp['Complex Name'] = tmp[['Definition', 'Complex Name']].apply(fn, axis = 1)

	fn = lambda x: '{:s}({:s})'.format(x['Gene Locus ID'], x['Complex ID'].split(':')[1].split('(')[0]) if ':' in x['Complex ID'] else '{:s}(1)'.format(x['Gene Locus ID'])
	tmp['Gene Locus ID'] = tmp[['Gene Locus ID', 'Complex ID']].apply(fn, axis = 1)

	fn = lambda x: '{:s}-MONOMER'.format(x['Gene Locus ID'].split('(')[0]) if x['Complex ID'] == '' else x['Complex ID'].split('(')[0].split(':')[0]
	tmp['Complex ID'] = tmp[['Gene Locus ID', 'Complex ID']].apply(fn, axis = 1)

	fn = lambda x: x[0] if isinstance(x, list) else x.tolist()[0]
	tmp = tmp.groupby(['Complex ID']).agg({'Gene Locus ID': lambda x: ' AND '.join(sorted(set(x.tolist()))), 'Complex Name': fn}).reset_index()
	tmp = tmp[['Complex ID', 'Complex Name', 'Gene Locus ID']].set_index('Complex ID')

	return tmp

def _get_df_ptms(df):
	tmp = df.copy(deep = True)
	#tmp = tmp[~tmp['Feature Type'].isin(['pseudo'])]
	tmp = tmp[tmp['Cofactors in Modified Complex'].notna()]

	if not tmp.empty:
		tmp['Complex ID'] = tmp['Complex ID'].apply(lambda x: x.split('(')[0].split(':')[0] if isinstance(x, str) else x)

		fn = lambda x: x['Complex ID'].split(':')[0] + ''.join([ '_mod_{:s}'.format(x) for x in x['Cofactors in Modified Complex'].split(' AND ')])
		tmp['Modified Complex'] = tmp[['Complex ID', 'Cofactors in Modified Complex']].apply(fn, axis = 1)

		# Modified complex names can be replaced using an alias
		# Copy-on-Write pandas 3.0 behavior is default to be True
		# tmp['Modified Complex'].update(tmp['Complex Name'].str.replace(' ', '_')) # inplace
		tmp.update({ 'Modified Complex' : tmp['Complex Name'].str.replace(' ', '_') })

		tmp = tmp[['Modified Complex', 'Complex ID', 'Cofactors in Modified Complex']]
		tmp.columns = ['Modified_enzyme', 'Core_enzyme', 'Modifications']
		tmp = tmp.drop_duplicates(keep = 'first').set_index('Modified_enzyme')

		return tmp
	else:
		return pandas.DataFrame(columns = ['Modified Complex', 'Complex ID', 'Cofactors in Modified Complex'])

def _get_df_enz2rxn(df, filter_in = set(), generics = False):
	tmp = df.copy(deep = True)
	#tmp = _correct_input(tmp)
	#tmp = tmp[~tmp['Feature Type'].isin(['pseudo'])]
	tmp = tmp[tmp['M-model Reaction ID'].notna()]

	tmp['Gene Locus ID'] = tmp['Gene Locus ID'].apply(lambda x: '{:s}-MONOMER'.format(x))
	tmp['Complex ID'] = tmp['Complex ID'].apply(lambda x: x.split(':')[0] if isinstance(x, str) else x)

	fn = lambda x: x['Complex ID'].split(':')[0] + ''.join([ '_mod_{:s}'.format(x) for x in x['Cofactors in Modified Complex'].split(' AND ')]) \
		if isinstance(x['Cofactors in Modified Complex'], str) else numpy.nan
	tmp['Modified Complex'] = tmp[['Complex ID', 'Cofactors in Modified Complex']].apply(fn, axis = 1)

	# collapse
	# Copy-on-Write pandas 3.0 behavior is default to be True
	#tmp['Modified Complex'].update(tmp['Generic Complex ID']) # inplace
	tmp.update({ 'Modified Complex' : tmp['Generic Complex ID'] })
	# tmp['Complex ID'].update(tmp['Modified Complex']) # inplace
	tmp.update({ 'Complex ID' : tmp['Modified Complex'] })
	# tmp['Gene Locus ID'].update(tmp['Complex ID']) # inplace
	tmp.update({ 'Gene Locus ID' : tmp['Complex ID'] })

	# Modified complex names can be replaced using an alias
	# Copy-on-Write pandas 3.0 behavior is default to be True
	# tmp['Gene Locus ID'].update(tmp['Complex Name'].str.replace(' ', '_')) # inplace
	tmp.update({ 'Gene Locus ID' : tmp['Complex Name'].str.replace(' ', '_') })

	tmp = tmp.groupby(['M-model Reaction ID']).agg({'Gene Locus ID': lambda x: ' OR '.join(sorted(set(x.tolist())))})
	tmp.columns = ['Complexes']

	return tmp

def _get_df_rna_enzs(df, filter_in = set(), generics = False):
	tmp = df[~df['Feature Type'].isin(['tRNA', 'rRNA'])]
	tmp = tmp[tmp['RNA mods/enzyme'].notna()]

	if not tmp.empty:
		tmp = _correct_input(tmp)
		tmp['RNA mods/enzyme'] = tmp['RNA mods/enzyme'].str.split(',')
		tmp = tmp.explode('RNA mods/enzyme')

		tmp['modification'] = tmp['RNA mods/enzyme'].apply(lambda x: x.split('_at_')[0])
		tmp['positions'] = tmp['RNA mods/enzyme'].apply(lambda x: x.split('_at_')[1])

		tmp['Gene Locus ID'] = tmp['Gene Locus ID'].apply(lambda x: '{:s}-MONOMER'.format(x))
		tmp['Complex ID'] = tmp['Complex ID'].apply(lambda x: x.split(':')[0] if isinstance(x, str) else x)

		fn = lambda x: x['Complex ID'].split(':')[0] + ''.join([ '_mod_{:s}'.format(x) for x in x['Cofactors in Modified Complex'].split(' AND ')]) \
			if isinstance(x['Cofactors in Modified Complex'], str) else numpy.nan
		tmp['Modified Complex'] = tmp[['Complex ID', 'Cofactors in Modified Complex']].apply(fn, axis = 1)

		# collapse
		# Copy-on-Write pandas 3.0 behavior is default to be True
		# tmp['Modified Complex'].update(tmp['Generic Complex ID']) # inplace
		tmp.update({ 'Modified Complex' : tmp['Generic Complex ID'] })
		# tmp['Complex ID'].update(tmp['Modified Complex']) # inplace
		tmp.update({ 'Complex ID' : tmp['Modified Complex'] })
		# tmp['Gene Locus ID'].update(tmp['Complex ID']) # inplace
		tmp.update({ 'Gene Locus ID' : tmp['Complex ID'] })

		tmp = tmp[['Gene Locus ID', 'modification', 'positions']]
		tmp.columns = ['enzymes', 'modification', 'positions']
		tmp = tmp.drop_duplicates(keep = 'first')

		return tmp.groupby(['modification', 'positions']).agg({'enzymes': lambda x: ' AND '.join(x.to_list())}).reset_index()
	else:
		return pandas.DataFrame(columns = ['enzymes', 'modification', 'positions'])

def _get_df_rna_ptms(df, filter_in = set(), generics = False):
	tmp = df[df['Feature Type'].isin(['tRNA', 'rRNA'])]
	tmp = tmp[tmp['RNA mods/enzyme'].notna()]

	if not tmp.empty:
		tmp['RNA mods/enzyme'] = tmp['RNA mods/enzyme'].str.split(',')
		tmp = tmp.explode('RNA mods/enzyme')

		tmp['modification'] = tmp['RNA mods/enzyme'].apply(lambda x: x.split('_at_')[0])
		tmp['positions'] = tmp['RNA mods/enzyme'].apply(lambda x: x.split('_at_')[1])

		# collapse
		# Copy-on-Write pandas 3.0 behavior is default to be True
		# tmp['Gene Locus ID'].update(tmp['Generic Complex ID']) # inplace
		tmp.update({ 'Gene Locus ID' : tmp['Generic Complex ID'] })

		tmp = tmp[['Gene Locus ID', 'modification', 'positions']]
		tmp.columns = ['bnum', 'modification', 'positions']
		tmp = tmp.drop_duplicates(keep = 'first')

		return tmp
	else:
		return pandas.DataFrame(columns = ['bnum', 'modification', 'positions'])

def _get_df_protloc(df, filter_in = set(), generics = False):
	tmp = df.copy(deep = True)
	#tmp = tmp[~tmp['Feature Type'].isin(['pseudo'])]
	tmp = tmp[tmp['Complex Location'].notna()]

	if not tmp.empty:
		tmp['Complex ID'] = tmp['Complex ID'].apply(lambda x: x.split(':')[0] if isinstance(x, str) else x)
		tmp['Monomer ID'] = tmp['Gene Locus ID'].apply(lambda x: '{:s}-MONOMER'.format(x))

		# collapse
		# Copy-on-Write pandas 3.0 behavior is default to be True
		# tmp['Monomer ID'].update(tmp['Complex ID']) # inplace
		tmp.update({ 'Monomer ID' : tmp['Complex ID'] })

		tmp = tmp[['Monomer ID', 'Complex Location', 'Gene Locus ID', 'Subunit Location', 'Translocation Pathway']]
		tmp.columns = ['Complex', 'Complex_compartment', 'Protein', 'Protein_compartment', 'translocase_pathway']
		tmp = tmp.drop_duplicates(subset = ['Complex', 'Protein', 'Protein_compartment'], keep = 'first').set_index('Complex')

		return tmp
	else:
		#return pandas.DataFrame(columns = ['Complex ID', 'Complex Location', 'Gene Locus ID', 'Subunit Location', 'Translocation Pathway'])
		return pandas.DataFrame(columns = ['Complex', 'Complex_compartment', 'Protein', 'Protein_compartment', 'translocase_pathway'])

def _get_df_transpaths(df, filter_in = set(), generics = False):
	tmp = df[~df['MetaComplex ID'].isnull()]
	tmp = tmp[tmp['MetaComplex ID'].str.match('translocation_pathway')]

	if not tmp.empty:
		tmp = _correct_input(tmp)
		tmp['Complex ID'] = tmp['Complex ID'].apply(lambda x: x.split(':')[0] if isinstance(x, str) else x)

		# collapse
		# Copy-on-Write pandas 3.0 behavior is default to be True
		# tmp['Complex ID'].update(tmp['Generic Complex ID'])
		tmp.update({ 'Complex ID' : tmp['Generic Complex ID'] })
		# tmp['Gene Locus ID'].update(tmp['Complex ID'])
		tmp.update({ 'Gene Locus ID' : tmp['Complex ID'] })

		return tmp[['Gene Locus ID', 'MetaComplex ID']].groupby('MetaComplex ID').agg(lambda x: list(set(x.tolist())))
	else:
		return pandas.DataFrame(columns = ['Gene Locus ID'], index = ['MetaComplex ID'])

def get_df_input_from_excel(df, df_rxns):
	lst = [
		#'Complex Name',
		#'Complex ID',
		'Cofactors in Modified Complex',
		'Generic Complex ID',
		'MetaComplex ID',
		'ME-model SubReaction',
		'M-model Reaction ID',
		'RNA mods/enzyme'
		]

	tmp1 = df[df['Feature Type'].str.match('CDS')]#.dropna(subset = lst, how = 'all', axis = 0)
	tmp2 = df[df['Feature Type'].isin(['rRNA', 'tRNA', 'ncRNA', 'tmRNA', 'misc_RNA', 'pseudo'])]
	df = pandas.concat([tmp1, tmp2], axis = 0).drop_duplicates()
	df = df.fillna({
		'Gene Locus ID' : '',
		'Reversibility' : numpy.nan,
		#'Spontaneous' : 'False', # see reactions.txt input file
		'GroEL_dependent_folding' : 'False',
		'DnaK_dependent_folding' : 'False',
		'N_terminal_methionine_cleavage' : 'False',
		'RNA stability' : 'False'
		})

	tmp = _get_df_rxns(df)
	for idx, x in df_rxns[df_rxns.index.isin(tmp.index)].iterrows():
		logging.warning('The reaction \'{:s}\' appears in the M-model and in the \'df_metadata_orphan_rxns\' input (default value)'.format(idx))
		logging.warning('If you want to use the M-model \'{:s}\' reaction, delete the ID from \'df_metadata_orphan_rxns\'. Otherwise, add the ID to \'defer_to_rxn_matrix\'.'.format(idx))

	# remove entries that are in df_rxns (user input overrides m-model info)
	tmp = tmp[~tmp.index.isin(df_rxns.index)]
	df_rxns = pandas.concat([tmp, df_rxns]).fillna('False')
	df_cplxs = _get_df_cplxs(df)
	df_ptms = _get_df_ptms(df)
	df_enz2rxn = _get_df_enz2rxn(df)
	df_protloc = _get_df_protloc(df)
	df_transpaths = _get_df_transpaths(df)

	# modifications of tRNAs and rRNAs
	df_rna_enzs = _get_df_rna_enzs(df)
	df_rna_ptms = _get_df_rna_ptms(df)
	cols = ['modification', 'positions']

	if not df_rna_ptms.empty or not df_rna_enzs.empty:
		df_rna_mods = pandas.merge(df_rna_ptms, df_rna_enzs, how = 'left', left_on = cols, right_on = cols).fillna('No_Machine')
	else:
		df_rna_mods = pandas.DataFrame(columns = ['bnum', 'modification', 'positions', 'enzymes'])

	return df, df_rxns, df_cplxs, df_ptms, df_enz2rxn, df_rna_mods, df_protloc, df_transpaths

def filter_out_cofactor(df, cofactors_to_skip):
	df = df.copy(deep = True)
	def filt(old_cofactors, *lst):
		if isinstance(old_cofactors, str):
			new_string = ' AND '.join([ cofactor for cofactor in old_cofactors.split(' AND ') if cofactor.split('(')[0] not in lst ])
			return new_string if new_string != '' else None
		else:
			return old_cofactors

	df['Cofactors in Modified Complex'] = df['Cofactors in Modified Complex'].apply(filt, args = (cofactors_to_skip))
	return df

def filter_in_cofactor(df, cofactors_to_keep):
	df = df.copy(deep = True)
	def filt(old_cofactors, *lst):
		if isinstance(old_cofactors, str):
			new_string = ' AND '.join([ cofactor for cofactor in old_cofactors.split(' AND ') if cofactor.split('(')[0] in lst ])
			return new_string if new_string != '' else None
		else:
			return old_cofactors

	df['Cofactors in Modified Complex'] = df['Cofactors in Modified Complex'].apply(filt, args = (cofactors_to_keep))
	return df



import tqdm
import json
import cobra
import pandas

import logging
log = logging.getLogger(__name__)

import coralme
from collections import defaultdict, Counter

def fix_id(id_str):
	return id_str.replace('_DASH_', '__')

def read(filename) -> pandas.DataFrame:
	return pandas.read_csv(filename, sep = '\t', index_col = None, comment = '#', skip_blank_lines = True, dtype = str, keep_default_na = False)

def get_complex_subunit_stoichiometry(complex_stoichiometry, rna_components = set()) -> dict:
	"""Returns dictionary of prot/prot, prot/rna complexes: {stoichiometry: {locus_tag: count}}
	"""
	complex_stoichiometry_dict = {}
	complex_stoichiometry.columns = ['Name', 'Stoichiometry']

	for key, row in complex_stoichiometry.iterrows():
		if key in complex_stoichiometry_dict.keys():
			logging.warning('Complex \'{:s}\' is present twice or more times in the DataFrame and the repeated item was ignored.'.format(key))
		else:
			complex_stoichiometry_dict[key] = {}

		for bnums in row['Stoichiometry'].split(' AND '):
			bnum, num = bnums.rstrip(')').split('(')
			stoichiometry = float(num) if not num == '' else 1.
			# add prefix to complex component
			#prefix = 'protein_' if bnum not in rna_components else 'RNA_'
			prefix = 'RNA_' if bnum in rna_components else 'protein_' if 'generic' not in bnum else ''
			complex_stoichiometry_dict[key][prefix + bnum] = stoichiometry

	return complex_stoichiometry_dict

def get_reaction_matrix_dict(matrix_df, compartments = {}, complex_set = set()) -> dict:
	"""Return dictionary representation of the metabolic reaction matrix.
	Updates metabolite id with compartment if not contained in complex_list
	"""
	#matrix_df.columns = ['Reaction', 'Metabolites', 'Compartment', 'Stoichiometry']
	matrix_df.columns = ['Reaction', 'Metabolites', 'Stoichiometry']
	#matrix_df.replace({'No_Compartment': 'Cytoplasm'}, inplace = True)

	# old code; replaced to get compartments from the me-model object
	#compartments = {'Cytosol': 'c', 'Periplasm': 'p', 'Extra-organism': 'e'}
	metabolic_reaction_dict = defaultdict(dict)
	for idx, row in matrix_df.iterrows():
		reaction = fix_id(row['Reaction'])
		metabolite = fix_id(row['Metabolites'])
		stoichiometry = row['Stoichiometry']
		#if compartments.get(row['Compartment']) is None:
			#compartment_id = ''
		#else:
			#compartment_id = '_{:s}'.format(compartments.get(row['Compartment']))

		# use compartment to append appropriate suffix
		#if metabolite.split('_mod_')[0] not in complex_set:
			#metabolite += compartment_id
		metabolic_reaction_dict[reaction][metabolite] = float(stoichiometry)

	return metabolic_reaction_dict

def get_complex_modifications(reaction_matrix, protein_complexes, complex_mods, compartments = set()) -> dict:
	"""
	"""
	#complex_dct = get_complex_subunit_stoichiometry(protein_complexes)
	#complex_set = set(complex_dct.keys())

	#ignored_complexes = set()
	#for met_stoich in coralme.builder.flat_files.get_reaction_matrix_dict(reaction_matrix).values():
		#for met, value in met_stoich.items():
			#if len(met.split('_mod_')[1:]) >= 1:
				#ignored_complexes.add(met)

	## correct ignored complexes
	#ignored_complexes = [ x for x in ignored_complexes
		#if x.split('_mod_')[1:] != ['pydx5p(1)'] and x.split('_mod_')[1:] != ['pydx5p(2)'] and x.split('_mod_')[1:] != ['4fe4s(2)']]

	new_mod_dict = {}
	for key, value in complex_mods.T.to_dict().items():
		# QueG_mono_mod_4fe4s(2), CPLX0-782_mod_4fe4s(2), CPLX0-246_CPLX0-1342_mod_pydx5p(1) and IscS_mod_pydx5p(2) must be exceptions
# 		if key in ignored_complexes:
# 			continue
		key = key.replace('_DASH_', '__')
		new_mod_dict[key] = {}
		new_mod_dict[key]['core_enzyme'] = value['Core_enzyme']
		new_mod_dict[key]['modifications'] = {}
		for mods in value['Modifications'].split(' AND '):
			mod, num_mods = mods.rstrip(')').split('(')
			if num_mods == '':
				num_mods = 1.
			else:
				num_mods = float(num_mods)

			mod = mod.replace('_DASH_', '__')
			new_mod_dict[key]['modifications'][mod + '_c'] = -num_mods

	return new_mod_dict

def get_reaction_to_complex(m_model, enz2rxn, modifications = True):
	"""anything not in this dict is assumed to be an orphan"""
	enz2rxn.columns = ['Complexes']
	enz2rxn = enz2rxn.apply(lambda x: x.replace('DASH', ''))

	rxn_to_complex_dict = defaultdict(set)
	for reaction, complexes in enz2rxn.itertuples():
		for cplx in complexes.split(' OR '):
			if modifications:
				rxn_to_complex_dict[reaction].add(cplx)
			else:
				rxn_to_complex_dict[reaction].add(cplx.split('_mod_')[0])

	for reaction in m_model.reactions:
		if 's0001' in reaction.gene_reaction_rule:
			rxn_to_complex_dict[reaction.id].add(None)

	return rxn_to_complex_dict

def remove_compartment(id_str):
	# original ecolime remove_compartment function
	#return id_str.replace('_c', '').replace('_p', '').replace('_e', '')
	return '_'.join(id_str.split('_')[:-1]) # compartment ID follows the last underscore

def process_reaction_matrix_dict(reaction_matrix, cplx_data, me_compartments = set()):
	reaction_matrix['Reaction'] = reaction_matrix['Reaction'].str.split(',')
	reaction_matrix = reaction_matrix.explode('Reaction')
	complex_dct = get_complex_subunit_stoichiometry(cplx_data)
	reaction_matrix_dict = get_reaction_matrix_dict(reaction_matrix, compartments = me_compartments, complex_set = set(complex_dct.keys()))

	return reaction_matrix_dict

def process_m_model(
	m_model, # mets_data, rxns_data, # all data
	mets_data, rxns_data, reaction_matrix, cplx_data, # only new data
	me_compartments = set(), defer_to_rxn_matrix = list(), repair = True):

	# copy the M-model object
	m_model = m_model.copy()

	# old code, known bug
	#for rxn in m_model.reactions:
		#if rxn.id.startswith('EX_') or rxn.id.startswith('DM_') or rxn.id.startswith('SK_') or rxn.id.startswith('BIOMASS_'):
			#continue

		## old code
		##if rxn.id not in reaction_matrix_dict.keys() or rxn.id in defer_to_rxn_matrix:
		#if rxn.id in defer_to_rxn_matrix:
			#rxn.remove_from_model(remove_orphans = True)
			#logging.warning('The MetabolicReaction \'{:s}\' (using \'defer_to_rxn_matrix\') was removed from the M-model metabolic network.'.format(rxn.id))

	#m_model.remove_reactions([ m_model.reactions.get_by_id(rxn) for rxn in defer_to_rxn_matrix ])
	rxns_to_remove = [ m_model.reactions.get_by_id(rxn) for rxn in defer_to_rxn_matrix if m_model.reactions.has_id(rxn) ]
	m_model.remove_reactions(rxns_to_remove)
# 	mets_to_remove = [ m for m in m_model.metabolites if len(m.reactions) == 0 ]
# 	m_model.remove_metabolites(mets_to_remove)

	# correct stoichiometry if rxn.check_mass_balance() == {'charge': +1.0, 'H': +1.0} or integer multiples of it
	count = 0
	for rxn in m_model.reactions:
		if rxn.id.startswith(('EX_', 'DM_', 'SK_')):
			continue

		check = rxn.check_mass_balance()
		if check == {}:
			continue
		elif set(check.keys()) == {'H', 'charge'} and check['charge'] == check['H']:
			compt = rxn.get_compartments()
			if len(compt) == 1:
				metabolites = Counter(m_model.reactions.get_by_id(rxn.id).metabolites)
				metabolites.update(Counter({ m_model.metabolites.get_by_id('h_{:s}'.format(compt[0])) : -1 * check['H'] }))

				m_model.reactions.get_by_id(rxn.id)._metabolites = { k:v for k,v in metabolites.items() if v != 0. }
				logging.warning('Stoichiometry for \'{:s}\' was corrected to mass balance protons.'.format(rxn.id))
			else:
				logging.warning('Stoichiometry for \'{:s}\' was not corrected due to more than one compartment detected in the reaction. Please check and correct if needed.'.format(rxn.id))
			count += 1
		else:
			logging.warning('Reaction \'{:s}\' is not mass/charge balanced. Please check and correct if needed.'.format(rxn.id))
			count += 1
	if count == 0:
		logging.warning('No mass/charge imbalance was detected in M-model (excluding prefixed EX, DM, and SK reactions).')

	# remove unused genes
	cobra.manipulation.delete.remove_genes(m_model, [ x for x in m_model.genes if len(x.reactions) == 0 ], remove_reactions = False)
	cobra.manipulation.delete.prune_unused_reactions(m_model) # reactions without metabolites
	# cobra.manipulation.delete.prune_unused_metabolites(m_model) # orphan metabolites

	# met_data DataFrame
	mets_data = mets_data[mets_data['type'].isin(['ADD', 'REPLACE'])]
	mets_data.columns = ['me_id', 'name', 'formula', 'compartment', 'type']
	# WARNING: Do we need to rename the index?
	#mets_data.rename(index = lambda x: x.replace('_DASH_', '__'), inplace = True)

	# process protein_complex DataFrame, and...
	complex_dct = get_complex_subunit_stoichiometry(cplx_data)
	# ...process reaction_matrix DataFrame
	reaction_matrix_dict = get_reaction_matrix_dict(reaction_matrix, compartments = me_compartments, complex_set = set(complex_dct.keys()))

	for rxn_id in reaction_matrix_dict:
		if rxn_id in m_model.reactions:
			logging.warning('The MetabolicReaction \'{:s}\' (using \'reaction_matrix.txt\') is already present in the M-model and it will be replaced.'.format(rxn_id))
			m_model.remove_reactions([rxn_id], remove_orphans = False)

		# Metabolites need to be added into the M-model first
		rxn_stoichiometry = reaction_matrix_dict[rxn_id]
		for met in rxn_stoichiometry:
			try:
				met_obj = m_model.metabolites.get_by_id(met)
			except KeyError:
				if met.startswith('RNA_'):
					met_obj = coralme.core.component.TranscribedGene(str(met), 'tRNA', '')
				elif met.startswith('generic_tRNA'):
					met_obj = coralme.core.component.GenerictRNA(str(met))
				else:
					met_obj = coralme.core.component.Metabolite(str(met))
				m_model.add_metabolites([met_obj])

			met_id = remove_compartment(met_obj.id)
			if met_id in mets_data.index and not met_obj.formula:
				met_obj.name = mets_data.loc[met_id, 'name']
				met_obj.formula = mets_data.loc[met_id, 'formula']

		# Add new reactions into the M-model
		rxn = coralme.core.reaction.MEReaction(rxn_id)
		m_model.add_reactions([rxn])
		rxn.add_metabolites(rxn_stoichiometry)
		if rxn_id in rxns_data.index:
			reversible = rxns_data.loc[rxn_id, 'is_reversible']
			subsystem = 'Not Determined' if rxns_data.loc[rxn_id, 'subsystems'] == 'False' else rxns_data.loc[rxn_id, 'subsystems']
		else:
			reversible = 'False'
			logging.warning('Unable to determine MetabolicReaction \'{:s}\' reversibility. Default value is \'False\'.'.format(rxn_id))
			subsystem = 'Not Determined'
			logging.warning('Unable to determine MetabolicReaction \'{:s}\' subsystem. Default value is \'Not Determined\'.'.format(rxn_id))

		rxn.lower_bound = -1000.0 if reversible.lower() == 'true' else 0.0
		logging.warning('The MetabolicReaction \'{:s}\' was created into the M-model (using \'reaction_matrix.txt\').'.format(rxn_id))
		logging.warning('MetabolicReaction \'{:s}\' reversibility is set to \'{:s}\' (using \'reaction_matrix.txt\').'.format(rxn_id, reversible))
		logging.warning('MetabolicReaction \'{:s}\' subsystem is set to \'{:s}\' (using \'reaction_matrix.txt\').'.format(rxn_id, subsystem))
		rxn.subsystem = subsystem

	# m_to_me_map DataFrame
	#m_to_me_map.columns = ['me_name']
	m_to_me_map = mets_data[mets_data['me_id'].notnull()]
	#m_to_me_map.columns = ['me_id', 'name', 'formula', 'compartment', 'data_source']
	m_to_me_map.rename(lambda x: x.replace('_DASH_', '__'), inplace = True)

	m_model.add_metabolites([ coralme.core.component.Complex(id = x) for x in set(m_to_me_map['me_id']) ])

	for rxn in m_model.reactions:
		#met_id = remove_compartment(met.id)
		# old code. mets_data contains all the metabolites (m+me)
		#if met_id not in mets_data.index and met_id in m_to_me_map.index:
		#if met_id in mets_data.index:
			# old code. m_to_me_map contains the new metabolite ID that maps the M-model metabolite
			#met_id = m_to_me_map.loc[met.id, 'me_name']
		#if mets_data.loc[met_id, 'me_id'] != '' and mets_data.loc[met_id, 'me_id'] != 'N/A':
		for met, coeff in rxn.metabolites.items():
			if met.id in m_to_me_map.index and m_to_me_map.loc[met.id, 'me_id'] != '':
				#met.id = mets_data.loc[met_id, 'me_id']
				#met_obj = coralme.core.component.Complex(id = m_to_me_map.loc[met.id, 'me_id'])
				#met_obj.name = m_to_me_map.loc[met.id, 'name']
				#met_obj.formula = m_to_me_map.loc[met.id, 'formula']

				new_id = m_model.metabolites.get_by_id(m_to_me_map.loc[met.id, 'me_id'])
				old_stoich = rxn.metabolites[m_model.metabolites.get_by_id(met.id)]
				rxn.add_metabolites({ new_id : old_stoich })
				logging.warning('Metabolite \'{:s}\' was replaced with \'{:s}\' in MetabolicReaction \'{:s}\'.'.format(met.id, m_to_me_map.loc[met.id, 'me_id'], rxn.id))

	m_model.remove_metabolites([ m_model.metabolites.get_by_id(x) for x in m_to_me_map[m_to_me_map['type'].str.fullmatch('REPLACE|REMOVE')].index if m_model.metabolites.has_id(x) ])

	# Add new metabolites (ME-metabolites) with properties into the "M-model"
	for m_met_id in m_to_me_map.index:
		me_met_id = m_to_me_map.loc[m_met_id, 'me_id']
		if m_model.metabolites.has_id(me_met_id):
			met_obj = m_model.metabolites.get_by_id(me_met_id)
		else:
			met_obj = coralme.core.component.Metabolite(str(me_met_id))

		met_obj.name = m_to_me_map.loc[m_met_id, 'name']
		met_obj.formula = m_to_me_map.loc[m_met_id, 'formula'] if m_to_me_map.loc[m_met_id, 'formula'] != '' else None
		met_obj.compartment = m_met_id.split('_')[-1]
		# WARNING: This "fails" silently if the met_obj ID already exists
		m_model.add_metabolites([met_obj])

	if repair:
		m_model.repair()

	return m_model

def get_m_model(
	modelID = 'M-model-from-ME-model',
	metabolites = 'metabolites.txt',
	reactions = 'reactions.txt',
	reaction_matrix = 'reaction_matrix.txt',
	prot_complexes = 'protein_complexes.txt',
	exchanges = 'exchange_bounds.txt',
	sources_and_sinks = 'reaction_matrix_sources_and_sinks.txt',
	compartments = {}
	) -> cobra.core.model.Model:

	m = cobra.core.model.Model(modelID)

	# match compartments in ME-model
	m.compartments = compartments
	compartment_lookup = {v:k for k,v in m.compartments.items()}

	met_info = pandas.read_csv(metabolites, sep = '\t', index_col = None, comment = '#', skip_blank_lines = True)
	met_info.columns = ['id', 'name', 'formula', 'compartment', 'data_source']
	met_info = met_info.set_index('id')

	complex_set = set(get_complex_subunit_stoichiometry(filename = prot_complexes).keys())

	for met_id in met_info.index:
		fixed_id = fix_id(met_id)
		for compartment in met_info.compartment[met_id].split('AND'):
			compartment = compartment.strip()
			if compartment == 'No_Compartment':
				print('Assigned {:s} to c'.format(met_id))
				compartment = m.compartments['c']
			new_met = cobra.Metabolite(
				fixed_id + '_' + compartment_lookup[compartment])
			new_met.name = met_info.name[met_id]
			new_met.formula = met_info.formula[met_id]
			m.add_metabolites(new_met)

	rxn_info = get_reaction_info_frame(reactions)
	rxn_dict = get_reaction_matrix_dict(filename = reaction_matrix, compartments = compartments, complex_set = complex_set)
	for rxn_id in rxn_info.index:
		reaction = coralme.core.reaction.MEReaction(rxn_id)
		reaction.name = rxn_info.description[rxn_id]
		for met_id, amount in rxn_dict[rxn_id].items():
			try:
				metabolite = m.metabolites.get_by_id(met_id)
			except KeyError:
				metabolite = cobra.Metabolite(met_id)
			reaction.add_metabolites({metabolite: amount})
		reaction.lower_bound = -1000. if rxn_info.is_reversible[rxn_id].lower() == 'true' else 0.
		reaction.upper_bound = 1000.
		if rxn_info.is_spontaneous[rxn_id].lower() == 'true':
			reaction.gene_reaction_rule = 's0001'
		m.add_reaction(reaction)

	sources_sinks = pandas.read_csv(sources_and_sinks, sep = '\t', index_col = None, comment = '#', skip_blank_lines = True)
	sources_sinks.columns = ['rxn_id', 'met_id', 'compartment', 'stoic']
	sources_sinks = sources_sinks.set_index('met_id')
	sources_sinks.index = [fix_id(i) for i in sources_sinks.index]

	source_amounts = pandas.read_csv(exchanges, sep = '\t', index_col = None, comment = '#', skip_blank_lines = True)
	source_amounts.columns = ['met_id', 'amount']
	source_amounts = source_amounts.set_index('met_id')
	source_amounts.index = [fix_id(i) for i in source_amounts.index]

	for met in sources_sinks.index:
		met_id = met + '_' + compartment_lookup[sources_sinks.compartment[met]]
		# EX_ or DM_ + met_id
		reaction_id = sources_sinks.rxn_id[met][:3] + met_id
		reaction = coralme.core.reaction.MEReaction(reaction_id)
		m.add_reaction(reaction)
		reaction.add_metabolites({m.metabolites.get_by_id(met_id): -1})
		# set bounds on exchanges
		if reaction.id.startswith('EX_') and met in source_amounts.index:
			reaction.lower_bound = -source_amounts.amount[met]

	return m

def get_trna_modification_targets(trna_mod) -> dict:
	trna_mod.columns = ['bnum', 'modification', 'positions', 'enzymes']

	trna_mod_dict = defaultdict(dict)
	for idx, mod in trna_mod.iterrows():
		mod_loc = '{:s}_at_{:s}'.format(mod['modification'], mod['positions'])
		trna_mod_dict[mod['bnum']][mod_loc] = 1.

	return trna_mod_dict

def get_reaction_keffs(me, keffs = 'keffs.json', verbose = True):
	def log(*args, **kwargs):
		if verbose:
			print(*args, **kwargs)

	with open(keffs, 'r') as infile:
		keffs = json.load(infile)

	new_keffs = {}
	for met_rxn in me.reactions:
		# skip spontaneous reactions
		if getattr(met_rxn, 'complex_data', None) is None:
			continue

		if isinstance(met_rxn, coralme.core.reaction.MetabolicReaction) and r.complex_data.id != 'CPLX_dummy':
			key = met_rxn.id.replace('-', '_DASH_').replace('__', '_DASH_').replace(':', '_COLON_')
			# key = met_rxn.id
			key = 'keff_' + key.replace('_FWD_', '_').replace('_REV_', '_')

			matches = [i for i in keffs if key in i]
			# get the direction
			if met_rxn.reverse:
				matches = [i for i in matches if i.endswith('_reverse_priming_keff')]
			else:
				matches = [i for i in matches if i.endswith('_forward_priming_keff')]

			if len(matches) == 1:
				new_keffs[met_rxn.id] = keffs[matches[0]]
			elif len(matches) > 0:
				if len(matches) == len([i for i in matches if key + '_mod_']):
					new_keffs[met_rxn.id] = keffs[matches[0]]
				else:
					log(key, len(matches))
			else:  # len(matches) == 0
				log('no keff found for ' + key)

	return new_keffs

def get_m_to_me_metabolite_mapping(filename = 'm_to_me_mets.txt', m_mets = 'm_name', me_mets = 'me_name') -> dict:
	"""returns a mapping from m metabolites to me metabolites"""
	df = pandas.read_csv(filename, sep = '\t', index_col = None, comment = '#', skip_blank_lines = True)
	return df.set_index(m_mets)[me_mets].dropna().to_dict()

def get_folding_data(filename, index_name = 'gene name', column_prefix = 'Dill_Keq_', column_suffix = 'C') -> pandas.DataFrame:
	"""returns formatted data length-based approximation of protein folding and aggregation data."""
	df = pandas.read_csv(filename, sep = '\t', index_col = None, comment = '#', skip_blank_lines = True)
	df = df.rename(columns = lambda x: x.replace(column_prefix, '').replace(column_suffix, ''))
	return df.set_index(index_name)



#!/usr/bin/python3
import re
import copy
import pandas
import tqdm
import cobra
import coralme
from warnings import warn


# from cobrame without changes
def get_base_complex_data(model, complex_id):
	"""
	If a complex is modified in a metabolic reaction it will not
	have a formation reaction associated with it. This function returns
	the complex data of the "base" complex, which will have the subunit
	stoichiometry of that complex
	"""

	# First try unmodified complex id
	try_1 = complex_id.split('_mod_')[0]
	if try_1 in model.process_data:
		return model.process_data.get_by_id(try_1)

	count = 0
	try_2 = complex_id.split('_mod_')[0] + '_'
	for i in model.process_data.query(try_2):
		if isinstance(i, coralme.core.processdata.ComplexData):
			count += 1
			data = i

	if count == 0:
		raise UserWarning('No base complex found for \'{:s}\'.'.format(complex_id))

	elif count > 1:
		raise UserWarning('More than one possible base complex found for \'{:s}\'.'.format(complex_id))

	return data

# from cobra with changes
def parse_composition(tmp_formula) -> dict:
	"""Break the chemical formula down by element."""
	element_re = re.compile("([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)")
	#tmp_formula = self.formula
	# commonly occuring characters in incorrectly constructed formulas
	if "*" in tmp_formula:
		warn(f"invalid character '*' found in formula '{self.formula}'")
		tmp_formula = self.formula.replace("*", "")
	if "(" in tmp_formula or ")" in tmp_formula:
		warn(f"parenthesis found in formula '{self.formula}'")
		return
	composition = {}
	parsed = element_re.findall(tmp_formula)
	for element, count in parsed:
		if count == "":
			count = 1
		else:
			try:
				count = float(count)
				int_count = int(count)
				if count == int_count:
					count = int_count
				else:
					warn(f"{count} is not an integer (in formula {self.formula})")
			except ValueError:
				warn(f"failed to parse {count} (in formula {self.formula})")
				#self.elements = {}
				return
		if element in composition:
			composition[element] += count
		else:
			composition[element] = count
	return composition

## Network traversing
def substitute_value(m,coeff):
	"""Substitute coefficient with a value"""
	if not hasattr(coeff,'subs'):
		return coeff
	return coeff.subs([(m._model.mu, 1e-6)])

def get_next_from_type(l,t):
	"""Get the first object that matches a type"""
	try:
		if isinstance(l,dict):
			return next(i for i,v in l.items() if isinstance(i,t) and substitute_value(i,v) > 0)
		return next(i for i in l if isinstance(i,t))
	except:
		return set()

def find_complexes(m, seen = set()):
	"""Find final complexes that are formed from a metabolite"""
	if not m:
		return set()
	if m in seen:
		return set()

	seen.add(m)

	# Metabolite objects
	if isinstance(m,coralme.core.component.TranslatedGene):
		cplxs = set()
		for r in m.reactions:
			if substitute_value(m,r.metabolites[m] > 0):
				continue
			cplxs = cplxs | find_complexes(r, seen=seen)
		return cplxs
	if isinstance(m,coralme.core.component.TranscribedGene):
		translated_protein = m.id.replace('RNA_','protein_')
		if translated_protein in m._model.metabolites:
			return find_complexes(m._model.metabolites.get_by_id(translated_protein), seen=seen)
		cplxs = set()
		for r in m.reactions:
			if substitute_value(m,r.metabolites[m] > 0):
				continue
			cplxs = cplxs | find_complexes(r, seen=seen)
		return cplxs
	if isinstance(m,coralme.core.component.ProcessedProtein):
		cplxs = set()
		for r in m.reactions:
			if substitute_value(m,r.metabolites[m] > 0):
				continue
			cplxs = cplxs | find_complexes(r, seen=seen)
		return cplxs

	if isinstance(m,coralme.core.component.Complex) or isinstance(m,coralme.core.component.GenericComponent) or isinstance(m,coralme.core.component.GenerictRNA):
		other_formations = [r for r in m.reactions if (isinstance(r,coralme.core.reaction.ComplexFormation) or isinstance(r,coralme.core.reaction.GenericFormationReaction)) and substitute_value(m,r.metabolites[m]) < 0]
#		 print(other_formations)
		cplxs = set([m])
		if other_formations:
			for r in other_formations:
				cplxs = cplxs | find_complexes(r, seen=seen)
		return cplxs
#		 return set([m])

	# Reaction objects
	if isinstance(m,coralme.core.reaction.PostTranslationReaction):
		return find_complexes(get_next_from_type(m.metabolites,coralme.core.component.ProcessedProtein), seen=seen)
	if isinstance(m,coralme.core.reaction.ComplexFormation):
		return find_complexes(get_next_from_type(m.metabolites,coralme.core.component.Complex), seen=seen)
	if isinstance(m,coralme.core.reaction.GenericFormationReaction):
		return find_complexes(get_next_from_type(m.metabolites,coralme.core.component.GenericComponent), seen=seen)
	if isinstance(m,coralme.core.reaction.tRNAChargingReaction):
		return find_complexes(get_next_from_type(m.metabolites,coralme.core.component.GenerictRNA), seen=seen)
	if isinstance(m,coralme.core.reaction.MetabolicReaction):
		return find_complexes(get_next_from_type(m.metabolites,coralme.core.component.Complex), seen=seen) | \
				find_complexes(get_next_from_type(m.metabolites,coralme.core.component.GenericComponent), seen=seen)
	return set()

def get_functions(cplx):
	"""Find final functions (Metabolic, Translation, ...) of a Complex."""
	allowed_types = [
		coralme.core.component.Complex,
		coralme.core.component.GenericComponent,
		coralme.core.component.GenerictRNA
		]
	assert any(isinstance(cplx,i) for i in allowed_types), f"Unexpected object of type {type(cplx)}"
	functions = set()
	for r in cplx.reactions:
		if isinstance(r,coralme.core.reaction.MetabolicReaction) and hasattr(r,'subsystem'):
			if r.subsystem:
				functions.add('Metabolic:' + r.subsystem)
				continue
			functions.add('Metabolic: No subsystem')
			continue
		if isinstance(r,coralme.core.reaction.TranslationReaction):
			functions.add('Translation')
		elif isinstance(r,coralme.core.reaction.TranscriptionReaction):
			functions.add('Transcription')
		elif isinstance(r,coralme.core.reaction.tRNAChargingReaction):
			functions.add('tRNA-Charging')
		elif isinstance(r,coralme.core.reaction.PostTranslationReaction):
			functions.add('Post-translation')
		elif isinstance(r,coralme.core.reaction.SummaryVariable):
			functions.add('Biomass')
	return functions

# Other
# Originally developed by JDTB, UCSD (2022)
def close_sink_and_solve(rxn_id):
	global _model
	global _muf
	model = copy.deepcopy(_model)
	model.reactions.get_by_id(rxn_id).bounds = (0, 0)
	model.optimize(max_mu = _muf, maxIter = 1)
	if not model.solution:
		return (rxn_id, False)
	else:
		return (rxn_id, True)

def change_reaction_id(model,old_id,new_id):
	"""Change the ID of a reaction"""
	old_rxn = model.reactions.get_by_id(old_id)
	rxn = cobra.Reaction(new_id)
	model.add_reactions([rxn])
	for k,v in old_rxn.metabolites.items():
		rxn.add_metabolites({k.id:v})
	rxn.bounds = old_rxn.bounds
	rxn.name = old_rxn.name
	rxn.subsystem = old_rxn.subsystem
	rxn.notes = old_rxn.notes
	rxn.gene_reaction_rule = old_rxn.gene_reaction_rule
	model.remove_reactions([old_rxn])

def get_metabolites_from_pattern(model,pattern):
	"""Get metabolites that contain a substring"""
	return [m.id for m in model.metabolites.query(pattern)]

def evaluate_lp_problem(Sf, Se, lb, ub, keys, atoms):
	"""Get an LP problem from NLP objects"""
	lb = [ x(*[ keys[x] for x in list(atoms) ]) if hasattr(x, '__call__') else float(x.xreplace(keys)) if hasattr(x, 'subs') else x for x in lb ]
	ub = [ x(*[ keys[x] for x in list(atoms) ]) if hasattr(x, '__call__') else float(x.xreplace(keys)) if hasattr(x, 'subs') else x for x in ub ]
	Se = { k:x(*[ keys[x] for x in list(atoms) ]) if hasattr(x, '__call__') else float(x.xreplace(keys)) if hasattr(x, 'subs') else x for k,x in Se.items() }

	Sf.update(Se)

	return Sf, Se, lb, ub

def is_producible(me,met,growth_key_and_value):
	"""Check if a metabolite is producible by the network"""
	if met not in me.metabolites:
		return False
	r = add_exchange_reactions(me, [met], prefix = 'DM_')[0]
	r.bounds = (1e-16,1e-16)
	if me.check_feasibility(keys = growth_key_and_value):
		r.remove_from_model()
		return True
	else:
		r.remove_from_model()
		return False


def get_transport_reactions(model,met_id,comps=['e','c']):
	"""Get transport reactions of a metabolite in a ME-model"""
	assert isinstance(model,coralme.core.model.MEModel), f"Unexpected model object of type {type(model)}"
	from_met = re.sub('_[a-z]$','_'+comps[0],met_id)
	to_met = re.sub('_[a-z]$','_'+comps[1],met_id)

	if isinstance(model,coralme.core.model.MEModel):
		reaction_type = ['MetabolicReaction']
	else:
		reaction_type = 0
	prod_rxns = [rxn.id for rxn in get_reactions_of_met(model,to_met,s=1,verbose=0,only_types=reaction_type)]
	cons_rxns = [rxn.id for rxn in get_reactions_of_met(model,from_met,s=-1,verbose=0,only_types=reaction_type)]
	transport_rxn_ids = list(set(prod_rxns)&set(cons_rxns))

	return [model.reactions.get_by_id(rxn_id) for rxn_id in transport_rxn_ids]

def get_all_transport_of_model(model):
	"""Get transport reactions in a ME-model"""
	assert isinstance(model,coralme.core.model.MEModel), f"Unexpected model object of type {type(model)}"
	transport_reactions = []
	for r in tqdm.tqdm(model.reactions):
		comps = r.get_compartments()
		if len(comps) > 1:
			transport_reactions.append(r.id)
	return list(set(transport_reactions))


def format_kcats_from_DLKcat(df):
	"""Formate the output from DLKcat to coralME-compatible"""
	# df = pandas.read_csv("./bacillus/building_data/bacillus_rxn_kcats.tsv",sep='\t',index_col=0).set_index("reaction")

	df2 = pandas.DataFrame(columns=["direction","complex","mods","keff"])
	for r,keff in df['Kcat value (1/s)'].items():
		d = {}
		rid,cplx = re.split("_FWD_|_REV_",r)
		base_cplx = cplx.split("_mod_")[0]
		mods = cplx.split(base_cplx)[1]
		mods = " AND ".join(mods.split("_mod_")[1:])
		direc = "FWD" if "FWD" in r else "REV"
		d[rid] = {
			"direction":direc,
			"complex":base_cplx,
			"mods":mods,
			"keff":keff
		}
		df2 = pandas.concat([df2,pandas.DataFrame.from_dict(d).T])
	df2.index.name = 'reaction'
	return df2
#	 df2.to_csv("./bacillus/building_data/keffs.txt",sep='\t')

# Flux analysis migration
# These reactions are being migrated to coralme.util.flux_analysis
def exchange_single_model(me, flux_dict = 0, solution=0):
	warn(
		"exchange_single_model will be removed soon."
		"Use coralme.util.flux_analysis.exchange_single_model instead",
		DeprecationWarning,
	)
	return coralme.util.flux_analysis.exchange_single_model(me,
														 flux_dict = flux_dict,
														 solution=solution)
def get_met_coeff(stoich,growth_rate,growth_key='mu'):
	warn(
		"get_met_coeff will be removed soon."
		"Use coralme.util.flux_analysis.get_met_coeff instead",
		DeprecationWarning,
	)
	return coralme.util.flux_analysis.get_met_coeff(stoich,
												 growth_rate,
												 growth_key=growth_key)

def summarize_reactions(model,met_id,only_types=(),ignore_types = ()):
	warn(
		"summarize_reactions will be removed soon."
		"Use coralme.util.flux_analysis.summarize_reactions instead",
		DeprecationWarning,
	)
	return coralme.util.flux_analysis.summarize_reactions(model,
													   met_id,
													   only_types=only_types,
													   ignore_types = ignore_types)
def flux_based_reactions(model,
						 met_id,
						 only_types=(),
						 ignore_types = (),
						 threshold = 0.,
						 flux_dict=0,
						 solution = None,
						 keffs=False,
						 verbose=False):
	warn(
		"flux_based_reactions will be removed soon."
		"Use coralme.util.flux_analysis.flux_based_reactions instead",
		DeprecationWarning,
	)
	return coralme.util.flux_analysis.flux_based_reactions(model,
						 met_id,
						 only_types=only_types,
						 ignore_types = ignore_types,
						 threshold = threshold,
						 flux_dict=flux_dict,
						 solution = solution,
						 keffs=keffs,
						 verbose=verbose)
def get_reactions_of_met(me,met,s = 0, ignore_types = (),only_types = (), verbose = False,growth_key='mu'):
	warn(
		"get_reactions_of_met will be removed soon."
		"Use coralme.util.flux_analysis.get_reactions_of_met instead",
		DeprecationWarning,
	)
	return coralme.util.flux_analysis.get_reactions_of_met(me,
														met,
														s = s,
														ignore_types = ignore_types,
														only_types = only_types,
														verbose = verbose,
														growth_key = growth_key)


def get_keffs_from_model(me):
	data = []
	for reaction in me.reactions:
		if hasattr(reaction, 'coupling_coefficient_enzyme') and 'dummy_reaction' not in reaction.id:
			data.append([reaction.id, reaction.keff])
	df = pandas.DataFrame(data, columns = ['reaction', 'keff'])

	tmp = df['reaction'].str.split('_mod_', expand = True, n = 1)
	df['reaction'] = tmp.iloc[:, 0].apply(lambda x: x.split('_FWD_')[0].split('_REV_')[0])
	df['complex'] = tmp.iloc[:, 0].apply(lambda x: x.split('_FWD_')[1] if 'FWD' in x else x.split('_REV_')[1])
	df['mods'] = tmp.iloc[:, 1].apply(lambda x: ' AND '.join(x.split('_mod_')) if x is not None else None)

	return df[['reaction', 'direction', 'complex', 'mods', 'keff']]

def check_me_coverage(builder):
	if not hasattr(builder, 'ref'):
		return NotImplemented
	if not hasattr(builder, 'df_data'):
		return NotImplemented

	ref_genes = [ x.id for x in builder.ref.m_model.genes ]
	tmp = builder.df_data[['Gene Locus ID', 'Reference BBH', 'M-model Reaction ID']]
	tmp = tmp[tmp['Reference BBH'].isin(ref_genes)]

	# implementation relies on 'M-model Reaction ID' is sorted before None
	tmp = tmp.sort_values(['Gene Locus ID', 'M-model Reaction ID']).drop_duplicates('Gene Locus ID', keep = 'first')
	res = tmp[tmp['M-model Reaction ID'].isna()]['Gene Locus ID'].values

	dct = builder.homology.mutual_hits
	for idx, gene in enumerate(res):
		res[idx] = '{:s} <-> {:s}'.format(res[idx], dct[gene])

	# store report to curation notes
	if len(res) > 0:
		builder.org.curation_notes['check_me_coverage'].append({
			'msg':'Some genes show homology, but no metabolic function.',
			'triggered_by':sorted(res),
			'importance':'critical',
			'to_do':'Manually curate the M-model to correct GPRs or incorporate new reactions.'})

	return len(res)



import copy
import sympy
import coralme
from collections import defaultdict

def stringify(element, number):
	number = int(number)
	return element if number == 1 else (element + str(number).rstrip('.'))

def elements_to_formula(obj, elements):
	#obj.formula = ''.join(stringify(e, n) for e, n in sorted(iteritems(elements)))
	obj.formula = ''.join(stringify(e, n) for e, n in sorted(elements.items()) if n != 0)

def eval_reaction_at_growth_rate(reaction, growth_rate):
	for key, value in reaction._metabolites.items():
		if isinstance(value, sympy.Basic):
			reaction._metabolites[key] = value.subs(coralme.util.mu, growth_rate)
		if isinstance(reaction, coralme.TranslationReaction) and (isinstance(key, coralme.Ribosome) or isinstance(key, coralme.TranscribedGene)):
			reaction._metabolites[key] = 0.
		if isinstance(reaction, coralme.TranscriptionReaction) and isinstance(key, coralme.RNAP):
			reaction._metabolites[key] = 0.
	return reaction

def get_elements_from_process_data(reaction, process_data, elements):
	"""
	If a modification is required to form a functioning macromolecule,
	update the element dictionary accordingly.
	"""

	#for sub, count in iteritems(process_data.subreactions):
	for sub, count in process_data.subreactions.items():
		sub_obj = reaction._model.process_data.get_by_id(sub)
		for e, n in sub_obj.element_contribution.items():
			elements[e] += n * count
	return elements

def check_transcription_mass_balance(reaction):
	transcription_data = reaction.transcription_data
	if len(transcription_data.excised_bases) > 0:
		return reaction.check_me_mass_balance()

	# Store and reset the formula strings for each RNA
	conserve_dict = {}
	for rna in transcription_data.RNA_products:
		rna_obj = reaction._model.metabolites.get_by_id(rna)
		conserve_dict[rna_obj] = rna_obj.formula
		rna_obj.formula = ''
		rna_obj.elements = {}

	elements = defaultdict(int)

	nucleotide_count = {
		#i.replace('tp_c', 'mp_c'): v for i, v in iteritems(transcription_data.nucleotide_count)
		i.replace('tp_c', 'mp_c'): v for i, v in transcription_data.nucleotide_count.items()
		}

	for nuc, value in nucleotide_count.items():
		nuc_obj = reaction._model.metabolites.get_by_id(nuc)
		for e, n in nuc_obj.elements.items():
			elements[e] += value * n

	# Remove -OH for each
	elements['H'] -= len(transcription_data.nucleotide_sequence)
	elements['O'] -= len(transcription_data.nucleotide_sequence)

	elements_to_formula(rna_obj, elements)
	mass_balance = reaction.check_me_mass_balance()

	for rna_obj, formula in conserve_dict.items():
		rna_obj.formula = formula

	return mass_balance

def check_me_model_mass_balance(model0):
	model = copy.deepcopy(model0)

	def should_skip(reaction):
		# Exchanges and demand reactions will not be mass balanced
		if reaction.id.startswith('DM_') or reaction.id.startswith('EX_'):
			return True
		# Generic metabolites do not have a formula and Dummy reaction will not
		# be mass balanced
		elif 'to_generic' in reaction.id or 'dummy_reaction' in reaction.id:
			return True
		# Global constraints do not have a formula (for summary variables)
		# tRNA charging reactions are mass balance checked by confirming
		# mass balance of each modification/subreaction
		elif isinstance(reaction, (coralme.SummaryVariable, coralme.tRNAChargingReaction)):
			return True
		else:
			return False

	# Set k_deg = 0 so model will be feasible at growth rate = 0
	model.global_info['k_deg'] = 0
	for r in model.reactions:
		if isinstance(r, coralme.TranslationReaction):
			r.update()

	output = {}
	for r in model.reactions:
		me_reaction = True
		if should_skip(r):
			continue
		if not isinstance(r, coralme.MEReaction) and r.check_mass_balance():
			output[r.id] = r.check_mass_balance()
			me_reaction = False
		elif not isinstance(r, coralme.MEReaction):
			me_reaction = False

		if me_reaction:
			# Lipoprotein metabolites can have different formulas based on
			# reaction used to synthesized them (TODO: How to handle this?)
			if 'lipid_modification' in r.id:
				r.update()
			# Check mass balance at growth rate = 0
			eval_reaction_at_growth_rate(r, 0.)
			# Transcription reactions must be handled differently to account
			# for multiple RNAs per TU
			if isinstance(r, coralme.TranscriptionReaction) and check_transcription_mass_balance(r):
				output[r.id] = check_transcription_mass_balance(r)
			elif not isinstance(r, coralme.TranscriptionReaction) and r.check_me_mass_balance():
				output[r.id] = r.check_me_mass_balance()

	# Each individual subreaction data should be mass balanced, if necessary
	for data in model.process_data:
		if hasattr(data, 'element_contribution'):
			# Skip trivial modifications (only one metabolite as reactant)
			# These are simple modifications
			if len(data.stoichiometry) == 1 and list(data.stoichiometry.values())[0] < 0:
				continue

			calculated_element_contribution = {}
			for key, value in data.calculate_element_contribution().items():
				if value != 0:
					calculated_element_contribution[key] = value

			set_element_contribution = {}
			for key, value in data._element_contribution.items():
				if value != 0:
					set_element_contribution[key] = value

			if calculated_element_contribution != set_element_contribution:
				txt = 'Calculated element contribution ({:s}) not equal to user defined element contribution({:s})'
				output[data.id] = txt.format(calculated_element_contribution, set_element_contribution)

	return output



import math
import numpy
import scipy

import coralme

# Experimental Data
#gr_data_doublings_per_hour = [0, 0.6, 1.0, 1.5, 2.0, 2.5]

# First point is mass 1 genome in 0.4 um^3 at density
#percent_dna_data = [0.0592, 0.0512, 0.0330, 0.0252, 0.0222, 0.0208]

def percent_dna_template_function(params, gr):
	[g_p_gdw_0, g_per_gdw_inf, b, d] = params
	c = g_per_gdw_inf
	a = g_p_gdw_0 - g_per_gdw_inf
	g_p_gdw = (-a * gr ** d) / (b + gr ** d) + a + c
	return g_p_gdw

def optimize_dna_function(gr, percent_dna):
	params = numpy.array([0.9, 0.3, 0.2, 1.0])

	def _minimization_function(params, gr, percent_dna):
		return percent_dna_template_function(params, gr) - percent_dna
	a = scipy.optimize.leastsq(_minimization_function, params, args = (gr, percent_dna))
	return a[0]

def get_dna_mw_no_ppi_dict(model):
	"""
	Return the molecular weight of dna component with the diphosphate group
	removed.
	"""
	default = {
		'dctp': 467.156923,
		'dgtp': 507.181023,
		'datp': 491.181623,
		'dttp': 482.168263,
		'ppi' : 177.97508200000001
		}

	dna_mw_no_ppi = {}
	ppi_mw = model.metabolites.ppi_c.formula_weight
	if ppi_mw == 0:
		ppi_mw = default['ppi']

	for dna in ['dctp', 'dgtp', 'datp', 'dttp']:
		dna_mw = model.metabolites.get_by_id(dna + '_c').formula_weight
		if dna_mw == 0:
			dna_mw = default[dna]

		dna_mw_no_ppi[dna] = dna_mw - ppi_mw

	return dna_mw_no_ppi

def return_gr_dependent_dna_demand(model, gc_fraction, percent_dna_data, gr_data_doublings_per_hour):
	"""
	Returns dNTP coefficients and lower/upper bounds of DNA_replication
	reaction
	"""
	if len(percent_dna_data) == len(gr_data_doublings_per_hour):
		pass
	else:
		raise Exception('The \'percent DNA data\' and \'growth data have different\' lengths.')

	gr_data = [m * math.log(2) for m in gr_data_doublings_per_hour]
	fit_params = optimize_dna_function(gr_data, percent_dna_data)
	dna_g_per_g = percent_dna_template_function(fit_params, model.mu)  # gDNA / gDW

	# average dinucleotide molecular weight
	dna_mw = get_dna_mw_no_ppi_dict(model)
	dntp_mw = (gc_fraction * (dna_mw['dctp'] + dna_mw['dgtp'])) / 2
	dntp_mw += ((1 - gc_fraction) * (dna_mw['datp'] + dna_mw['dttp'])) / 2

	# 1 / (gDNA / mol) * (1000 mmol / 1 mol)
	mmol_dntps_per_gram_dna = 1 / dntp_mw * 1000

	# (mmol / gDNA) * (gDNA / gDW)
	mmol_dntp_per_gdw = dna_g_per_g * mmol_dntps_per_gram_dna

	# lower and upper bound
	mmol_dntp_per_gdw_per_hr = mmol_dntp_per_gdw * model.mu

	return mmol_dntp_per_gdw_per_hr




ribosome_stoich = {'30_S_assembly':{'stoich':{'generic_16s_rRNAs':1}},
                    '50_S_assembly':{'stoich':{'generic_23s_rRNAs':1,'generic_5s_rRNAs':1}},
                    'assemble_ribosome_subunits': {'stoich': {'gtp_c': 1}}}

ribosome_subreactions = {
    'gtp_bound_30S_assembly_factor_phase1':
                         {'enzyme': '',
                          'stoich': {"gtp_c": -2,
                                     "h2o_c": -2,
                                     "h_c": 2,
                                     "pi_c": 2,
                                     "gdp_c": 2},
                          'num_mods': 1},
                         'RbfA_mono_assembly_factor_phase1':
                         {'enzyme': '',
                          'stoich': {},
                          'num_mods': 1},
                         'RimM_mono_assembly_factor_phase1':
                         {'enzyme': '',
                          'stoich': {},
                          'num_mods': 1}
}

generics = {'generic_16Sm4Cm1402': {'enzymes':[]},
                        'generic_LYSINEaaRS': {'enzymes':[]},
                        'generic_Dus': {'enzymes':[]},
                        'generic_RF': {'enzymes':[]},
                        'generic_Tuf': {'enzymes':[]},
                        'generic_RNase': {'enzymes':[]},
                        'generic_16s_rRNAs': {'enzymes':[]},
                        'generic_23s_rRNAs': {'enzymes':[]},
                        'generic_5s_rRNAs': {'enzymes':[]},
                        'generic_5s_rRNAs': {'enzymes':[]},
                        'generic_fes_transfers_complex': {'enzymes':[]}
           }

rrna_modifications = {
                      # ---------16S Modifications---------------
                      'm2G_at_1207': {'machine': '',  
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm2G_at_1516': {'machine': '',
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm2G_at_966': {'machine': '',  
                                     'metabolites': {'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 1}},
                      'm3U_at_1498': {'machine': '',  
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm4Cm_at_1402': {'machine': '',
                                       'metabolites': {'amet_c': -2,
                                                       'ahcys_c': 2,
                                                       'h_c': 2}},
                      'm5C_at_1407': {'machine': '',  
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm5C_at_967': {'machine': '',  
                                     'metabolites': {'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 1}},
                      'm62A_at_1518': {'machine': '',  
                                       'metabolites': {'amet_c': -2,
                                                       'ahcys_c': 2,
                                                       'h_c': 2}},
                      'm62A_at_1519': {'machine': '',  
                                       'metabolites': {'amet_c': -2,
                                                       'ahcys_c': 2,
                                                       'h_c': 2}},
                      'm7G_at_527': {'machine': '',  
                                     'metabolites': {'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 1}},
                      'Y_at_516': {'machine': '',  
                                   'metabolites': {}},

                      # ---------23S Modifications---------------
                      'Cm_at_2498': {'machine': '',
                                     'metabolites': {'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 1}},
                      'D_at_2449': {'machine': '',
                                    'metabolites': {'h_c': -1,
                                                    'nadh_c': -1,
                                                    'nad_c': 1}},
                      'Gm_at_2251': {'machine': '',  
                                     'metabolites': {'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 1}},
                      'm1G_at_745': {'machine': '',
                                     'metabolites': {'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 1}},
                      'm2A_at_2503': {'machine': '',
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm2G_at_1835': {'machine': '',  
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm2G_at_2445': {'machine': '',  
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm5C_at_1962': {'machine': '',  
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm5U_at_1939': {'machine': '',
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm5U_at_747': {'machine': '',
                                     'metabolites': {'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 1}},
                      'm6A_at_1618': {'machine': '',  
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm6A_at_2030': {'machine': '',
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'm7G_at_2069': {'machine': '',
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'Um_at_2552': {'machine': '',  
                                     'metabolites': {'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 1}},
                      'Y_at_1911': {'machine': '',
                                    'metabolites': {}},
                      'Y_at_1915': {'machine': '',
                                    'metabolites': {}},
                      'm3Y_at_1915': {'machine': '',  
                                      'metabolites': {'amet_c': -1,
                                                      'ahcys_c': 1,
                                                      'h_c': 1}},
                      'Y_at_1917': {'machine': '',
                                    'metabolites': {}},
                      'Y_at_2457': {'machine': '',  
                                    'metabolites': {}},
                      'Y_at_2504': {'machine': '',  
                                    'metabolites': {}},
                      'Y_at_2580': {'machine': '',  
                                    'metabolites': {}},
                      'Y_at_2604': {'machine': '',  
                                    'metabolites': {}},
                      'Y_at_2605': {'machine': '',  
                                    'metabolites': {}},
                      'Y_at_746': {'machine': '',  
                                   'metabolites': {}},
                      'Y_at_955': {'machine': '',  
                                   'metabolites': {}}}

rrna_modification_info = {'Y': {'elements': {}, 'charge': 0},
                     'm3Y': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'Um': {'elements': {'H': 2, 'C': 1}, 'charge': 0},
                     'm7G': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'm6A': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'm5U': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'm5C': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'm2G': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'm2A': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'm1G': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'Gm': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'D': {'elements': {'H': 2}, 'charge': 0},
                     'Cm': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'm62A': {'elements': {'C': 2, 'H': 4}, 'charge': 0},
                     'm4Cm': {'elements': {'C': 2, 'H': 4}, 'charge': 0},
                     'm3U': {'elements': {'C': 1, 'H': 2}, 'charge': 0}
                     }

amino_acid_trna_synthetase = {
  "cys__L_c": '',
  "leu__L_c": '',
  "lys__L_c": '',
  "asp__L_c": '',
  "phe__L_c": '',
  "his__L_c": '',
  "asn__L_c": '',
  "pro__L_c": '',
  "ala__L_c": '',
  "ile__L_c": '',
  "ser__L_c": '',
  "arg__L_c": '',
  "met__L_c": '',
  "tyr__L_c": '',
  "glu__L_c": '',
  "thr__L_c": '',
  "val__L_c": '',
  "gly_c": '',
  "trp__L_c": '',
  "gln__L_c": '',
}

amino_acid_regex = {'cys__L_c': 'cystein',
 'leu__L_c': '(?<!iso)leuc',
 'lys__L_c': 'lys',
 'asp__L_c': 'aspart',
 'phe__L_c': 'phenylalan',
 'his__L_c': 'histid',
 'asn__L_c': 'asparag',
 'pro__L_c': 'prol',
 'ala__L_c': 'alan',
 'ile__L_c': 'isoleuc',
 'ser__L_c': 'ser',
 'arg__L_c': 'argin',
 'met__L_c': 'methion',
 'tyr__L_c': 'tyros',
 'glu__L_c': 'glutam(?!inyl|ine)',
 'thr__L_c': 'threon',
 'val__L_c': 'val',
 'gly_c': 'glyc',
 'trp__L_c': 'tryptophan',
 'gln__L_c': 'glutamin'}

translation_stop_dict = {'UAG': {'enzyme':''},
                         'UGA': {'enzyme':''},
                         'UAA': {'enzyme':'generic_RF'}}

initiation_subreactions = {
    'Translation_initiation_factor_InfA':
        {'enzymes': [],
         'stoich': {},
        'element_contribution': {}},

    'Translation_initiation_factor_InfC':
        {'enzymes': [],
         'stoich': {},
        'element_contribution': {}},

    'Translation_gtp_initiation_factor_InfB':
        {'enzymes': [],
         'stoich': {'gtp_c': -1,
                    'h2o_c': -1,
                    'h_c': 1,
                    'pi_c': 1,
                    'gdp_c': 1},
        'element_contribution': {}},

    'fmet_addition_at_START':
        {'enzymes': [],
         # iOL had h_c:1 for fmet addition but this is not mass balanced
         'stoich': {'10fthf_c': -1, 'thf_c': 1,
                    # 'h_c': 1,
                    'generic_tRNA_START_met__L_c': -1},
         'element_contribution': {'C': 1, 'O': 1}}
   }

elongation_subreactions = {'FusA_mono_elongation': {'enzymes': [],
                                                    'stoich': {'gtp_c': -1,
                                                               'h2o_c': -1,
                                                               'h_c': 1,
                                                               'pi_c': 1,
                                                               'gdp_c': 1}},

                           'Tuf_gtp_regeneration': {'enzymes': [],
                                                    'stoich': {}}}
termination_subreactions = {'PrfA_mono_mediated_termination':
                            {'enzymes': [],
                             'stoich': {}},

                            'PrfB_mono_mediated_termination':
                            {'enzymes': [],
                             'stoich': {}},

                            'generic_RF_mediated_termination':
                            {'enzymes': [],
                             'stoich': {}},

                            'N_terminal_methionine_cleavage':
                            {'enzymes': [],
                             'stoich': {'h2o_c': -1,
                                        'met__L_c': 1, 'h_c': 1},
                             'element_contribution': {'H': -10, 'O': -1,
                                                      'C': -5, 'N': -1,
                                                      'S': -1}},

                            'peptide_deformylase_processing':
                            {'enzymes': [],
                             'stoich': {'h2o_c': -1,
                                        'for_c': 1},
                             'element_contribution':
                                 {'H': 1, 'O': -1, 'C': -1}},

                            # This is a GTPS
                            'peptide_chain_release':
                            {'enzymes': [],
                             'stoich': {'gtp_c': -1,
                                        'h2o_c': -1,
                                        'h_c': 1,
                                        'pi_c': 1,
                                        'gdp_c': 1}},

                            'ribosome_recycler':
                            {'enzymes': [],
                             'stoich': {}},

                            'GroEL_dependent_folding':
                            {'enzymes': [],
                             'stoich': {'atp_c': -7,
                                        'h2o_c': -7,
                                        'h_c': 7,
                                        'adp_c': 7,
                                        'pi_c': 7}},

                            # DnaK is correct
                            'DnaK_dependent_folding':
                            {'enzymes': [],
                             'stoich': {'atp_c': -1,
                                        'h2o_c': -1,
                                        'h_c': 1,
                                        'adp_c': 1,
                                        'pi_c': 1}}
                            }
special_trna_subreactions = {
    'sec_addition_at_UGA': {
        'enzymes': [],  # Selenocysteine loaders
        'stoich': {'h_c': 1, 'selnp_c': -1,
                   'pi_c': 1,
                   'generic_tRNA_UGA_ser__L_c': -1}, # Serine is the precursor to Sec
        'element_contribution': {'O': -1, 'Se': 1}}}

excision_machinery = {
    'rRNA_containing': {'enzymes':[]},
    'monocistronic': {'enzymes':[]},
    'polycistronic_wout_rRNA': {'enzymes':[]}}

folding_dict = {
    "GroEL_dependent_folding": {'enzymes':[]},
    "DnaK_dependent_folding": {'enzymes':[]}
}

special_modifications = {
    'fes_transfer':{'enzymes':[],'stoich':{}},
    'fes_chaperones':{'enzymes':[],'stoich':{}},
    'generic_2fe2s_transfer_complex':{'enzymes':[],'stoich':{}},
    'generic_4fe4s_transfer_complex':{'enzymes':[],'stoich':{}},
    'lipoate_modifications':{'enzymes':[],'stoich':{}},
    'bmocogdp_chaperones':{'enzymes':[],'stoich':{}}
}

trna_modification = {'D_at_20A': {'enzymes': [],
                                  'stoich': {'nadph_c': -1,
                                                  'h_c': -1,
                                                  'nadp_c': 1}},

                     'D_at_20': {'enzymes': [],  
                                 'stoich': {'nadph_c': -1,
                                                 'h_c': -1,
                                                 'nadp_c': 1}},

                     't6A_at_37': {'enzymes': [],  
                                   'stoich': {'hco3_c': -1,
                                                   'thr__L_c': -1,
                                                   'atp_c': -1,
                                                   'amp_c': 1,
                                                   'h_c': 1,
                                                   'h2o_c': 1,
                                                   'ppi_c': 1}},

                     'm7G_at_46': {'enzymes': [],  
                                   'stoich': {'amet_c': -1,
                                                   'ahcys_c': 1,
                                                   'h_c': 1}},

                     'acp3U_at_47': {'enzymes': [],
                                     'stoich': {'amet_c': -1,
                                                     '5mta_c': 1,
                                                     'h_c': 1}},

                     'm5U_at_54': {'enzymes': [],  
                                   'stoich': {'amet_c': -1,
                                                   'ahcys_c': 1,
                                                   'h_c': 1}},

                     'Y_at_55': {'enzymes': [],  
                                 'stoich': {}},

                     'Y_at_65': {'enzymes': [],  
                                 'stoich': {}},

                     'D_at_17': {'enzymes': [],  
                                 'stoich': {'nadph_c': -1,
                                                 'h_c': -1,
                                                 'nadp_c': 1}},

                     'cmo5U_at_34': {'enzymes': [],
                                     'stoich': {'amet_c': -1,
                                                     'chor_c': -2,
                                                     'ahcys_c': 1,
                                                     'h_c': 1,
                                                     'C10H8O5_c': 1,
                                                     'C9H9O4_c': 1}},

                     'D_at_16': {'enzymes': [],  
                                 'stoich': {'nadph_c': -1,
                                                 'h_c': -1,
                                                 'nadp_c': 1}},

                     'Q_at_34': {
                     'enzymes': [],
                     'stoich': {'preq1_c': -1,
                                     'amet_c': -1,
                                     'gua_c': 1,
                                     'ade_c': 1,
                                     'met__L_c': 1,
                                     'h_c': 2}},

                     'm2A_at_37': {'enzymes': [],
                                   'stoich': {'amet_c': -1,
                                                   'ahcys_c': 1,
                                                   'h_c': 1}},

                     's4U_at_8': {'enzymes': [],
                                  'carriers': {},
                                  'stoich': {'atp_c': -1,
                                                  'amp_c': 1,
                                                  'ppi_c': 1,
                                                  'h_c': 1}},

                     'm6t6A_at_37': {'enzymes': [],
                                     'stoich': {'amet_c': -1,
                                                     'atp_c': -1,
                                                     'hco3_c': -1,
                                                     'thr__L_c': -1,
                                                     'ahcys_c': 1,
                                                     'amp_c': 1,
                                                     'h_c': 2,
                                                     'h2o_c': 1,
                                                     'ppi_c': 1}},
                     's2C_at_32': {'enzymes': [],
                                   'carriers': {},
                                   'stoich': {'atp_c': -1,
                                                   'amp_c': 1,
                                                   'ppi_c': 1,
                                                   'h_c': 1}},
                     'mnm5U_at_34': {'enzymes': [],
                                     'stoich': {'gtp_c': -1,
                                                     'h2o_c': -1,
                                                     '5fthf_c': -1,
                                                     'gly_c': -1,
                                                     'amet_c': -1,
                                                     'ahcys_c': 1,
                                                     'h_c': 3,
                                                     'gdp_c': 1,
                                                     'glx_c': 1,
                                                     'pi_c': 1,
                                                     'thf_c': 1}},

                     'Y_at_40': {'enzymes': [],  
                                 'stoich': {}},

                     'Gm_at_18': {'enzymes': [],  
                                  'stoich': {'amet_c': -1,
                                                  'ahcys_c': 1,
                                                  'h_c': 1}},

                     'Um_at_32': {'enzymes': [],
                                  'stoich': {'amet_c': -1,
                                                  'ahcys_c': 1,
                                                  'h_c': 1}},

                     'Y_at_38': {'enzymes': [],  
                                 'stoich': {}},

                     'ac4C_at_34': {'enzymes': [],  
                                    'stoich': {'accoa_c': -1,
                                                    'coa_c': 1}},

                     'Y_at_39': {'enzymes': [],  
                                 'stoich': {}},
                     
                     'mnm5s2U_at_34': {'enzymes': [],
                                       'carriers': {},
                                       'stoich': {'atp_c': -1,
                                                       'gtp_c': -1,
                                                       'h2o_c': -1,
                                                       '5fthf_c': -1,
                                                       'gly_c': -1,
                                                       'amet_c': -1,
                                                       'gdp_c': 1,
                                                       'pi_c': 1,
                                                       'h_c': 4,
                                                       'thf_c': 1,
                                                       'glx_c': 1,
                                                       'ahcys_c': 1,
                                                       'amp_c': 1,
                                                       'ppi_c': 1}},

                     'm6A_at_37': {'enzymes': [],
                                   'stoich': {'amet_c': -1,
                                                   'ahcys_c': 1,
                                                   'h_c': 1}},

                     'Cm_at_32': {'enzymes': [],
                                  'stoich': {'amet_c': -1,
                                                  'ahcys_c': 1,
                                                  'h_c': 1}},

                     'ms2i6A_at_37': {'enzymes': [],

                                      'carriers': {},
                                      'stoich': {'dmpp_c': -1,
                                                      'amet_c': -2,
                                                      'ppi_c': 1,
                                                      'ahcys_c': 1,
                                                      'h_c': 2,
                                                      'met__L_c': 1,
                                                      'dad__5_c': 1,
                                                      }},

                     'Y_at_32': {'enzymes': [],  
                                 'stoich': {}},

                     'D_at_21': {'enzymes': [],  
                                 'stoich': {'nadph_c': -1,
                                                 'h_c': -1,
                                                 'nadp_c': 1}},

                     'm1G_at_37': {'enzymes': [],  
                                   'stoich': {'amet_c': -1,
                                                   'ahcys_c': 1,
                                                   'h_c': 1}},

                     'Y_at_13': {'enzymes': [],  
                                 'stoich': {}},

                     'k2C_at_34': {'enzymes': [],  
                                   'stoich': {'atp_c': -1,
                                                   'lys__L_c': -1,
                                                   'ppi_c': 1,
                                                   'amp_c': 1,
                                                   'h_c': 2}},

                     'I_at_34': {'enzymes': [],
                                 'stoich': {'h2o_c': -1, 'h_c': -1,
                                                 'nh4_c': 1}},

                     'i6A_at_37': {'enzymes': [],  
                                   'stoich': {'dmpp_c': -1,
                                                   'ppi_c': 1}},

                     'D_at_20_in_met_tRNA': {'enzymes': [],
                                             
                                             'stoich': {'nadph_c': -1,
                                                             'h_c': -1,
                                                             'nadp_c': 1}},

                     'D_at_16_in_met_tRNA': {'enzymes': [],
                                             
                                             'stoich': {'nadph_c': -1,
                                                             'h_c': -1,
                                                             'nadp_c': 1}},

                     'D_at_17_in_met_tRNA': {'enzymes': [],
                                             
                                             'stoich': {'nadph_c': -1,
                                                             'h_c': -1,
                                                             'nadp_c': 1}},
                     'D_at_20A_in_met_tRNA': {'enzymes': [],
                                              
                                              'stoich': {'nadph_c': -1,
                                                              'h_c': -1,
                                                              'nadp_c': 1}}
                     }

trna_modification_info = {'D': {'elements': {'H': 2}, 'charge': 0},
                     'i6A': {'elements': {'C': 5, 'H': 8}, 'charge': 0},
                     'I': {'elements': {'N': -1, 'H': -1, 'O': 1},
                           'charge': 0},
                     'k2C': {'elements': {'O': 1, 'N': 2, 'H': 12, 'C': 6},
                             'charge': 0},
                     'Y': {'elements': {}, 'charge': 0},
                     'm1G': {'elements': {'H': 2, 'C': 1}, 'charge': 0},
                     'ms2i6A': {'elements': {'C': 6, 'H': 10, 'S': 1},
                                'charge': 0},
                     'Cm': {'elements': {'H': 2, 'C': 1}, 'charge': 0},
                     'Um': {'elements': {'H': 2, 'C': 1}, 'charge': 0},
                     'm6A': {'elements': {'H': 2, 'C': 1}, 'charge': 0},
                     'mnm5s2U': {'elements': {'C': 2, 'H': 5, 'N': 1, 'O': -1,
                                              'S': 1},
                                 'charge': 0},
                     'ac4C': {'elements': {'H': 2, 'C': 2, 'O': 1},
                              'charge': 0},
                     'Gm': {'elements': {'H': 2, 'C': 1}, 'charge': 0},
                     'mnm5U': {'elements': {'C': 2, 'H': 5, 'N': 1},
                               'charge': 0},
                     's2C': {'elements': {'O': -1, 'S': 1}, 'charge': 0},
                     'm6t6A': {'elements': {'C': 6, 'O': 4, 'N': 1, 'H': 9},
                               'charge': 0},
                     's4U': {'elements': {'O': -1, 'S': 1}, 'charge': 0},
                     'm2A': {'elements': {'H': 2, 'C': 1}, 'charge': 0},
                     'Q': {'elements': {'C': 7, 'O': 2, 'H': 11}, 'charge': 1},
                     'cmo5U': {'elements': {'C': 2, 'O': 3, 'H': 2},
                               'charge': 0},
                     'm5U': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     'acp3U': {'elements': {'C': 4, 'H': 7, 'N': 1, 'O': 2},
                               'charge': 0},
                     'm7G': {'elements': {'C': 1, 'H': 2}, 'charge': 0},
                     't6A': {'elements': {'C': 5, 'N': 1, 'O': 4, 'H': 6},
                             'charge': 0}
                     }

peptide_processing_subreactions = {"peptide_deformylase_processing",
                                   "peptide_chain_release",
                                   "ribosome_recycler"}

translation_start_codons = {"AUG", "GUG", "UUG", "AUU", "CUG"}

transcription_subreactions = {
    'Transcription_normal_rho_independent':
        {'enzymes': [],
         'stoich': {}},
    'Transcription_normal_rho_dependent':
        {'enzymes': [],
         'stoich': {'atp_c': -3,
                    'h2o_c': -3,
                    'adp_c': 3,
                    'pi_c': 3,
                    'h_c': 3}},
    'Transcription_stable_rho_independent':
        {'enzymes': [],
         'stoich': {}},
    'Transcription_stable_rho_dependent':
        {'enzymes': [],
         'stoich': {'atp_c': -3,
                    'h2o_c': -3,
                    'adp_c': 3,
                    'pi_c': 3,
                    'h_c': 3}}
}

abbreviation_to_pathway = {'s': 'sec_translocation',
                           't': ['tat_translocation', 'tat_translocation_alt'],
                           'b': 'bam_translocation',
                           'l': 'lol_translocation',
                           'y': 'yidC_translocation',
                           'a': 'secA_translocation',
                           'p': 'srp_yidC_translocation',
                           'r': 'srp_translocation'}

pathway_to_abbreviation = {}
for a,p in abbreviation_to_pathway.items():
    if isinstance(p,list):
        for i in p:
            pathway_to_abbreviation[i.replace('_translocation','')] = a
    else:
        pathway_to_abbreviation[p.replace('_translocation','')] = a


#!/usr/bin/python3
import cobra
import glob
import pandas
import re
import tqdm
from Bio import SeqIO

# Written originally by Rodrigo Santibanez
def get_orthofinder_results(path, save = False):
	# index: Orthogroups; columns: organisms; data: single string, comma separated
	data = pandas.read_csv(path, sep = '\t', index_col = 0)

	# convert NaN to strings -> split string -> get unique IDs -> convert to list -> collapse columns into one
	fn = lambda x: list(set(str(x).split(', ')))
	tmp = data.map(fn).sum(axis = 1)

	# remove NaN and duplicates from lists
	fn = lambda x: [ y for y in list(set(x)) if y != 'nan' ]
	tmp = tmp.apply(fn)

	# explode lists
	tmp = tmp.reset_index(drop = False).explode(0)

	# save
	tmp = tmp[[0, 'Orthogroup']].sort_values([0, 'Orthogroup']).reset_index(drop = True)
	if save:
		tmp.to_csv('01.proteinID-to-orthogroupID.txt', sep = '\t', index = False, header = None)
	return tmp

def _get_genbank_features_as_dataframe(genbank_path):
	contigs = []

	for contig in SeqIO.parse(genbank_path, 'genbank'):
		contigs.append(contig)

	dct = {}
	for contig in contigs:
		for feature in contig.features:
			if feature.type == 'CDS':
				# if feature.qualifiers.get('protein_id', [None])[0] is not None:
				#     if feature.qualifiers.get('old_locus_tag', None) is not None:
				protID = feature.qualifiers.get('protein_id', [None])[0]
				if protID is not None:
					dct[protID] = feature.qualifiers.get('locus_tag', [None])[0]
				else:
					dct[feature.qualifiers.get('locus_tag', [None])[0]] = feature.qualifiers.get('locus_tag')[0]

	return pandas.DataFrame([dct]).T

def get_features_from_genbanks(reference_genbank_path : str = None, genbanks_target_directory : list = []):
	"""
	This assumes gene IDs in M-model are contained in protein_id/locus_tag only
	"""
	data = [_get_genbank_features_as_dataframe(reference_genbank_path)]

	files = sorted(glob.glob(genbanks_target_directory + '/*'))
	for genbank in files:
		data.append(_get_genbank_features_as_dataframe(genbank))

	tmp = pandas.concat(data, axis = 1)
	tmp.columns = ['Original'] + [ x.split('/')[1].split('.gb')[0] for x in files ]
	return tmp

def _get_homologues(x):
	if any(isinstance(x, list) for x in x.dropna().tolist()):
		return [ x for y in x.dropna().tolist() for x in y ]
	else:
		return x.dropna().tolist()

def map_features_between_assemblies(mapping, genbank_features, save = False):
	mapping = mapping.groupby('Orthogroup').agg(lambda x: x.tolist())

	data = []
	for idx, value in tqdm.tqdm(mapping.iterrows(), total = mapping.shape[0]):
		tmp = genbank_features[genbank_features.index.isin([ x.replace('kb|', '') for x in value.tolist()[0]])].agg(lambda x: _get_homologues(x))

		if type(tmp) is pandas.core.series.Series:
			tmp = tmp.to_frame().T

		# tmp.index = [idx] # why?!
		data.append(tmp)

	tmp = pandas.concat(data)
	if save:
		tmp.to_csv('02.locustagID-to-locustagID.txt', sep = '\t', index = False)
	return tmp

def map_genes_between_models(model, mapping_features, save = False):
	model = cobra.io.load_json_model(model)
	genes = [ x.id for x in model.genes ]

	def fn(x, gene):
		if isinstance(x['Original'], list) and gene in x['Original']:
			return True
		if isinstance(x['Original'], str) and gene == x['Original']:
			return True
		return False

	data = []
	found_genes = []
	missing_genes = []

	for gene in tqdm.tqdm(genes):
		tmp = mapping_features[mapping_features.apply(lambda x: fn(x, gene), axis = 1)]
		data.append(tmp)

		shape = tmp.shape
		if shape[0] == 0:
			missing_genes.append(gene)
		else:
			found_genes.append(gene)

	tmp = pandas.concat(data)#.drop('Original', axis = 1)
	tmp = tmp.explode('Original')
	tmp = tmp.map(lambda x: ' or '.join(x) if type(x) == list else x).drop_duplicates()
	if save:
		tmp.to_csv('03.mapping-refmodel-to-others.txt', sep = '\t')

	#tmp5 = tmp2[~tmp2['Original'].astype(str).str.contains('|'.join(genes)) & tmp2['Original'].notna()]
	#tmp5 = tmp5.drop('Original', axis = 1)
	#tmp5 = tmp5.map(lambda x: ' or '.join(x) if type(x) == list else x).drop_duplicates()
	#tmp5.to_csv('05.mapping-of-genes-not-in-Mmodel.txt', sep = '\t')

	return tmp, found_genes, missing_genes

def convert_gene_IDs(target, ref_model, mapping_genes, do_not_convert_gprs = ['rtranscription', 'dreplication', 'pbiosynthesis']):
	deleted = []
	model = cobra.io.load_json_model(ref_model)
	genes = [ x.id for x in model.genes ]

	dct = { re.compile(r'\b{:s}\b'.format(row['Original'])):row[target] for idx, row in mapping_genes[['Original', target]].iterrows() }
	dct

	## Modify IDs for consistency with genbank
	for rxn in list(model.reactions):
		if rxn.id in do_not_convert_gprs:
			continue
		if rxn.gene_reaction_rule != '':
			for key, value in dct.items():
				if value == '':
					value = 'NULL'
				if rxn.gene_reaction_rule == value:
					continue
				if len(re.findall(key, rxn.gene_reaction_rule)) == 0:
					continue
				try:
					new_gpr = re.sub(key, value, rxn.gene_reaction_rule).replace('(NULL or ', '(').replace('NULL or ', '').replace(' or NULL', '')
					rxn.gene_reaction_rule = new_gpr
				except TypeError:
					print(rxn.id, new_gpr)

	for rxn in list(model.reactions):
		rxn.update_genes_from_gpr()

	# remove reactions with "and NULL" <- pseudogenes associated to the model
	for rxn in list(model.reactions):
		if rxn.id in do_not_convert_gprs:
			continue
		if 'NULL' in rxn.gene_reaction_rule:
			# rxn.remove_from_model()
			deleted.append((rxn.id, rxn.gene_reaction_rule))
			try:
				new_gpr = rxn.gene_reaction_rule.replace('(NULL and', '(').replace(' and NULL', '').replace('NULL and ', '').replace('NULL', '')
				rxn.gene_reaction_rule = new_gpr
			except TypeError:
				print(rxn.id, new_gpr)

	# update_genes_from_gpr adds genes to model.genes
	cobra.manipulation.remove_genes(model, [g for g in model.genes if not g.reactions])
	cobra.io.save_json_model(model, 'New-M-models/m_model_{:s}.json'.format(target))
	del model
	return (target, deleted)

def run_all(orthogroups, ref_model, new_model_ids, ref_genbank, genbanks_directory):
	mapping_orthologues = get_orthofinder_results(orthogroups)
	genbank_features = get_features_from_genbanks(reference_genbank_path = ref_genbank, genbanks_target_directory = genbanks_directory)
	mapping_features = map_features_between_assemblies(mapping_orthologues, genbank_features)
	mapping_genes, genes, missing = map_genes_between_models(ref_model, mapping_features)

	res = []
	for idx in new_model_ids:
		if idx in mapping_genes.columns[1:]: # 'Original' should be position 0
			res.append(convert_gene_IDs(idx, ref_model, mapping_genes))
	return res

if __name__ == '__main__':
	# to be completed
	run_all()
